"""Zemax OpticStudio analyses from the Wavefront category."""

from __future__ import annotations

import os
import re
from tempfile import mkstemp
from typing import Literal

import numpy as np
import pandas as pd

import zospy.api.config as _config
from zospy.analyses.old.base import AnalysisResult, AttrDict, OnComplete, new_analysis
from zospy.api import constants
from zospy.utils.pyutils import atox
from zospy.utils.zputils import standardize_sampling, unpack_datagrid
from zospy.zpcore import OpticStudioSystem


def wavefront_map(
    oss: OpticStudioSystem,
    field: int = 1,
    surface: Literal["Image"] | int = "Image",
    wavelength: int = 1,
    show_as: constants.Analysis.ShowAs | str = "Surface",
    rotation: constants.Analysis.Settings.Rotations | str = "Rotate_0",
    sampling: constants.Analysis.SampleSizes | str = "64x64",
    polarization: constants.Analysis.Settings.Polarizations | str | None = None,
    reference_to_primary: bool = False,
    use_exit_pupil: bool = True,
    remove_tilt: bool = False,
    scale: float = 1.0,
    subaperture_x: float = 0.0,
    subaperture_y: float = 0.0,
    subaperture_r: float = 1.0,
    contour_format: str = "",
    oncomplete: OnComplete | str = OnComplete.Close,
) -> AnalysisResult:
    """Wavefront map analysis.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    field :
        The field that is to be analyzed. Defaults to 1.
    surface  str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    wavelength : int
        The wavelength number to use for the analysis. Defaults to 1.
    show_as : constants.Analysis.ShowAs | str
        Defines the output plot format. Defaults to 'Surface'.
    rotation : constants.Analysis.Settings.Rotations | str
        The rotation or surface plots for viewing. Defaults to 'Rotate_0'.
    sampling : constants.Analysis.SampleSizes | str
        The sampling. Defaults to '64x64'.
    polarization : constants.Analysis.Settings.Polarizations | str | None
        The polarization that is accounted for. When set to None, polarization is ignored. Defaults to None.
    reference_to_primary : bool
        Defines whether the aberrations are referenced to the reference sphere for the used wavelength or for the
        primary wavelength. If True, the reference sphere for the primary wavelength is used. Defaults to False.
    use_exit_pupil : bool
        Defines whether the exit pupil shape is used. Defaults to True.
    remove_tilt : bool
        Defines whether linear x- and y-tilt is removed from the data. Defaults to False.
    scale : float
        The scale factor for surface plots. Defaults to False.
    subaperture_x : float
        The subaperture x. Defaults to 0.0
    subaperture_y : float
        The subaperture t. Defaults to 0.0
    subaperture_r : float
        The subaperture r. Defaults to 1.0
    contour_format : str
        The contour format. Only used when show-As is set to 'Contour'. If set to an empty string, OpticStdio ignores
        it. Defaults to ''.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.

    Returns
    -------
    AnalysisResult
        A wavefront map analysis result.
    """
    analysis_type = constants.Analysis.AnalysisIDM.WavefrontMap

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    # Adjust settings
    analysis.set_field(field)
    analysis.set_surface(surface)
    analysis.set_wavelength(wavelength)
    analysis.Settings.ShowAs = constants.process_constant(constants.Analysis.ShowAs, show_as)
    analysis.Settings.Rotation = constants.process_constant(constants.Analysis.Settings.Rotations, rotation)
    analysis.Settings.Sampling = constants.process_constant(
        constants.Analysis.SampleSizes, standardize_sampling(sampling)
    )
    analysis.Settings.Polarization = constants.process_constant(constants.Analysis.Settings.Polarizations, polarization)
    analysis.Settings.ReferenceToPrimary = reference_to_primary
    analysis.Settings.UseExitPupil = use_exit_pupil
    analysis.Settings.RemoveTilt = remove_tilt
    analysis.Settings.Scale = scale
    analysis.Settings.Subaperture_X = subaperture_x
    analysis.Settings.Subaperture_Y = subaperture_y
    analysis.Settings.Subaperture_R = subaperture_r
    analysis.Settings.ContourFormat = contour_format

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)
    settings.loc["Field"] = analysis.get_field()
    settings.loc["Surface"] = analysis.Settings.Surface.GetSurfaceNumber()
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["ShowAs"] = str(analysis.Settings.ShowAs)
    settings.loc["Rotation"] = str(analysis.Settings.Rotation)
    settings.loc["Sampling"] = str(analysis.Settings.Sampling)
    settings.loc["Polarization"] = str(analysis.Settings.Polarization)
    settings.loc["ReferenceToPrimary"] = analysis.Settings.ReferenceToPrimary
    settings.loc["UseExitPupil"] = analysis.Settings.UseExitPupil
    settings.loc["RemoveTilt"] = analysis.Settings.RemoveTilt
    settings.loc["Scale"] = analysis.Settings.Scale
    settings.loc["Subaperture_X"] = analysis.Settings.Subaperture_X
    settings.loc["Subaperture_Y"] = analysis.Settings.Subaperture_Y
    settings.loc["Subaperture_R"] = analysis.Settings.Subaperture_R
    settings.loc["ContourFormat"] = analysis.Settings.ContourFormat

    data = []
    for ii in range(analysis.Results.NumberOfDataGrids):
        data.append(unpack_datagrid(analysis.Results.DataGrids[ii],
                                    # pass "center" and None for consistency with older ZOSPy versions
                                     cell_origin="center", label_rounding=None))

    if len(data) == 0:
        data = pd.DataFrame()
    elif len(data) == 1:
        data = data[0]
    else:
        data = pd.concat(data, axis=1)

    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def _structure_zernike_standard_coefficients_result(line_list: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Structures the result of a zernike standard coefficients analysis.

    Parameters
    ----------
    line_list : list of str
        The line list obtained by reading in the results

    Returns
    -------
    (pd.DataFrame, pd.DataFrame)
        Two dataframes, respectively the general results and the coefficients
    """
    zlinepat = re.compile(r"^Z\s+\d+")
    valuepat_start = re.compile(r"^((-)?\d+\{}\d+)".format(_config.DECIMAL_POINT))
    valuepat_any = re.compile(r"((-)?\d+(\{}\d+)?)".format(_config.DECIMAL_POINT))

    zernike_lines = [line for line in line_list if (zlinepat.search(line) is not None)]

    general_lines = [line for line in line_list if line not in zernike_lines]
    general_arr = [" ".join(line.split()) for line in general_lines if ":" in line]
    zernike_arr = np.array([line.replace(" ", "").split() for line in zernike_lines])
    general_data = pd.DataFrame(columns=["Value", "Unit"])
    for line in general_arr:
        spl = line.split(":", 1)
        if len(spl) == 1:
            ind = "".join([item.title() for item in spl[0].split()])
            general_data.loc[ind] = ["", ""]
        else:
            ind = "".join([item.title() for item in spl[0].split()])
            nvals = len(valuepat_any.findall(spl[1]))
            dat = spl[1].strip().split(maxsplit=nvals)
            if len(dat) == 0:
                val = ""
                unit = ""
                general_data.loc[ind] = [val, unit]
            elif len(dat) == 1:
                if ind == "Date":
                    val = dat[0]
                elif valuepat_start.search(dat[0]):  # value is number
                    val = atox(dat[0], float)
                else:
                    val = dat[0]
                unit = ""
                general_data.loc[ind] = [val, unit]
            elif len(dat) == 2:
                if valuepat_start.search(dat[0]):  # value is number
                    val = atox(dat[0], float)
                else:
                    val = dat[0]
                unit = dat[1]
                general_data.loc[ind] = [val, unit]
            else:
                for ii in range(len(dat) - 1):
                    if valuepat_start.search(dat[ii]):  # value is number
                        val = atox(valuepat_start.search(dat[ii]).group(), float)
                    else:
                        val = dat[ii]
                    unit = dat[-1]
                    general_data.loc["{}_{}".format(ind, ii)] = [val, unit]
    zernike_data = pd.DataFrame(index=zernike_arr[:, 0].copy(), columns=["Value", "Unit", "Function"])
    zernike_data.loc[zernike_arr[:, 0], "Value"] = list(map(lambda s: atox(s, float), zernike_arr[:, 1].copy()))
    zernike_data.loc[zernike_arr[:, 0], "Unit"] = "waves"
    zernike_data.loc[zernike_arr[:, 0], "Function"] = zernike_arr[:, 3].copy()

    return general_data, zernike_data


def zernike_standard_coefficients(
    oss: OpticStudioSystem,
    sampling: str = "64x64",
    maximum_term: int = 37,
    wavelength: str | int = 1,
    field: str | int = 1,
    reference_opd_to_vertex: bool = False,
    surface: str | int = "Image",
    sx: float = 0.0,
    sy: float = 0.0,
    sr: float = 0.0,
    oncomplete: OnComplete | str = OnComplete.Close,
    txtoutfile: str = None,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Zernike Standard Coefficient Analysis.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    sampling : str | int
        The sampling, either string (e.g. '64x64') or int. The integer will be treated as a ZOSAPI Constants integer.
    maximum_term : int
        The maximum Zernike term, defaults to 37.
    wavelength : int
        The wavelength number that is to be used. Defaults to 1 (the first wavelength).
    field:
        The field that is to be analyzed. Defaults to 1.
    reference_opd_to_vertex : bool
        If the OPD should be referenced to vertex. Defaults to False.
    surface : str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    sx : float
        The sx.
    sy : float
        The sy.
    sr : float
        The sr.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.
    txtoutfile : str | None
        The textfile to which the OpticStudio output is saved. If None, a temporary file will be created and deleted. If
        string, it should be a full system path with '.txt' as extension, after which the file will be saved in that
        location. Defaults to None.

    Returns
    -------
    AnalysisResult
        A ZernikeStandardCoefficients analysis result. Next to the standard data, the raw text return obtained from the
        analysis will be present under 'RawTextData' and the txtoutfile under 'TxtOutFile'.
    """
    analysis_type = constants.Analysis.AnalysisIDM.ZernikeStandardCoefficients

    if txtoutfile is None:
        fd, txtoutfile = mkstemp(suffix=".txt", prefix="zospy_")
        os.close(fd)
        cleantxt = True
    else:
        if not txtoutfile.endswith(".txt"):
            raise ValueError('txtfile should end with ".txt"')
        cleantxt = False

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    # Apply settings
    analysis.Settings.SampleSize = getattr(constants.Analysis.SampleSizes, standardize_sampling(sampling))
    analysis.Settings.MaximumNumberOfTerms = maximum_term
    analysis.set_wavelength(wavelength)
    analysis.set_field(field)
    analysis.Settings.ReferenceOBDToVertex = reference_opd_to_vertex  # ToDo: Monitor name with zemax updates
    analysis.set_surface(surface)
    analysis.Settings.Sx = sx
    analysis.Settings.Sy = sy
    analysis.Settings.Sr = sr

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Get results
    analysis.Results.GetTextFile(txtoutfile)
    line_list = [line for line in open(txtoutfile, "r", encoding=oss.ZOS.get_txtfile_encoding())]

    general_data, zernike_data = _structure_zernike_standard_coefficients_result(line_list)
    data = AttrDict(GeneralData=general_data, Coefficients=zernike_data)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Manually create settings as they cannot be accessed
    settings = pd.Series(name="Settings", dtype=object)

    settings.loc["SampleSize"] = str(analysis.Settings.SampleSize)
    settings.loc["MaximumNumberOfTerms"] = analysis.Settings.MaximumNumberOfTerms
    settings.loc["Wavelength"] = analysis.Settings.Wavelength.GetWavelengthNumber()  # Todo Evaluate with 'all'
    settings.loc["Field"] = analysis.Settings.Field.GetFieldNumber()  # Todo Evaluate with 'all'
    settings.loc["ReferenceOBDToVertex"] = analysis.Settings.ReferenceOBDToVertex
    settings.loc["Wavelength"] = analysis.Settings.Wavelength.GetWavelengthNumber()  # Todo Evaluate with 'all'
    settings.loc["Field"] = analysis.Settings.Field.GetFieldNumber()  # Todo Evaluate with 'all'
    settings.loc["ReferenceOBDToVertex"] = analysis.Settings.Surface.GetSurfaceNumber()  # Todo Evaluate with 'all'
    settings.loc["Sx"] = sx
    settings.loc["Sy"] = sy
    settings.loc["Sr"] = sr

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
        RawTextData=line_list,
        TxtOutFile=txtoutfile,
    )  # Set additional params

    # cleanup
    if cleantxt:
        os.remove(txtoutfile)

    return analysis.complete(oncomplete, result)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("DataType", "FoucaultShowAs", "KnifeType", "Types")

class DataType:
    Computed: DataType = None
    Reference: DataType = None
    Difference: DataType = None

class FoucaultShowAs:
    Surface: FoucaultShowAs = None
    Contour: FoucaultShowAs = None
    GreyScale: FoucaultShowAs = None
    InverseGreyScale: FoucaultShowAs = None
    FalseColor: FoucaultShowAs = None
    InverseFalseColor: FoucaultShowAs = None
    X_CrossSection: FoucaultShowAs = None
    Y_CrossSection: FoucaultShowAs = None

class KnifeType:
    Horiz_Above: KnifeType = None
    Horiz_Below: KnifeType = None
    Vert_Right: KnifeType = None
    Vert_Left: KnifeType = None

class Types:
    Linear: Types = None
    Log_Minus_3: Types = None
    Log_Minus_6: Types = None
    Log_Minus_9: Types = None
    Log_Minus_12: Types = None
    Log_Minus_15: Types = None
    Log_Minus_18: Types = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("FftPsfType", "PsfRotation", "PsfSampling")

class FftPsfType:
    Linear: FftPsfType = None
    Log: FftPsfType = None
    Phase: FftPsfType = None
    Real: FftPsfType = None
    Imaginary: FftPsfType = None

class PsfRotation:
    CW0: PsfRotation = None
    CW90: PsfRotation = None
    CW180: PsfRotation = None
    CW270: PsfRotation = None

class PsfSampling:
    PsfS_32x32: PsfSampling = None
    PsfS_64x64: PsfSampling = None
    PsfS_128x128: PsfSampling = None
    PsfS_256x256: PsfSampling = None
    PsfS_512x512: PsfSampling = None
    PsfS_1024x1024: PsfSampling = None
    PsfS_2048x2048: PsfSampling = None
    PsfS_4096x4096: PsfSampling = None
    PsfS_8192x8192: PsfSampling = None
    PsfS_16384x16384: PsfSampling = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("SettingsDataType", "ZemaxColor", "ZemaxOpacity")

class SettingsDataType:
    None_: SettingsDataType = None
    Integer: SettingsDataType = None
    Double: SettingsDataType = None
    Float: SettingsDataType = None
    String: SettingsDataType = None
    Byte: SettingsDataType = None
    Boolean: SettingsDataType = None

class ZemaxColor:
    Default: ZemaxColor = None
    Color1: ZemaxColor = None
    Color2: ZemaxColor = None
    Color3: ZemaxColor = None
    Color4: ZemaxColor = None
    Color5: ZemaxColor = None
    Color6: ZemaxColor = None
    Color7: ZemaxColor = None
    Color8: ZemaxColor = None
    Color9: ZemaxColor = None
    Color10: ZemaxColor = None
    Color11: ZemaxColor = None
    Color12: ZemaxColor = None
    Color13: ZemaxColor = None
    Color14: ZemaxColor = None
    Color15: ZemaxColor = None
    Color16: ZemaxColor = None
    Color17: ZemaxColor = None
    Color18: ZemaxColor = None
    Color19: ZemaxColor = None
    Color20: ZemaxColor = None
    Color21: ZemaxColor = None
    Color22: ZemaxColor = None
    Color23: ZemaxColor = None
    Color24: ZemaxColor = None
    NoColor: ZemaxColor = None

class ZemaxOpacity:
    P100: ZemaxOpacity = None
    P90: ZemaxOpacity = None
    P80: ZemaxOpacity = None
    P70: ZemaxOpacity = None
    P60: ZemaxOpacity = None
    P50: ZemaxOpacity = None
    P40: ZemaxOpacity = None
    P30: ZemaxOpacity = None
    P20: ZemaxOpacity = None
    P10: ZemaxOpacity = None
    P00: ZemaxOpacity = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("ToleranceColumn", "ToleranceOperandType")

class ToleranceColumn:
    Param1: ToleranceColumn = None
    Param2: ToleranceColumn = None
    Param3: ToleranceColumn = None
    Nominal: ToleranceColumn = None
    Min: ToleranceColumn = None
    Max: ToleranceColumn = None
    Comment: ToleranceColumn = None

class ToleranceOperandType:
    TOFF: ToleranceOperandType = None
    TRAD: ToleranceOperandType = None
    TCUR: ToleranceOperandType = None
    TFRN: ToleranceOperandType = None
    TTHI: ToleranceOperandType = None
    TCON: ToleranceOperandType = None
    TSDI: ToleranceOperandType = None
    TSDX: ToleranceOperandType = None
    TSDY: ToleranceOperandType = None
    TSDR: ToleranceOperandType = None
    TSTX: ToleranceOperandType = None
    TSTY: ToleranceOperandType = None
    TIRX: ToleranceOperandType = None
    TIRY: ToleranceOperandType = None
    TIRR: ToleranceOperandType = None
    TPAR: ToleranceOperandType = None
    TEDV: ToleranceOperandType = None
    TEDX: ToleranceOperandType = None
    TEDY: ToleranceOperandType = None
    TEDR: ToleranceOperandType = None
    TETX: ToleranceOperandType = None
    TETY: ToleranceOperandType = None
    TETZ: ToleranceOperandType = None
    TUDX: ToleranceOperandType = None
    TUDY: ToleranceOperandType = None
    TUTX: ToleranceOperandType = None
    TUTY: ToleranceOperandType = None
    TUTZ: ToleranceOperandType = None
    TIND: ToleranceOperandType = None
    TABB: ToleranceOperandType = None
    TEXI: ToleranceOperandType = None
    SAVE: ToleranceOperandType = None
    COMP: ToleranceOperandType = None
    CPAR: ToleranceOperandType = None
    CEDV: ToleranceOperandType = None
    CMCO: ToleranceOperandType = None
    STAT: ToleranceOperandType = None
    TWAV: ToleranceOperandType = None
    TNPS: ToleranceOperandType = None
    TNPA: ToleranceOperandType = None
    TMCO: ToleranceOperandType = None
    TEZI: ToleranceOperandType = None
    TCMU: ToleranceOperandType = None
    SEED: ToleranceOperandType = None
    COMM: ToleranceOperandType = None
    TPAI: ToleranceOperandType = None
    TCIO: ToleranceOperandType = None
    TCEO: ToleranceOperandType = None
    TNMA: ToleranceOperandType = None
    CNPA: ToleranceOperandType = None
    CNPS: ToleranceOperandType = None
    TRLX: ToleranceOperandType = None
    TRLY: ToleranceOperandType = None
    TRLR: ToleranceOperandType = None
    TARX: ToleranceOperandType = None
    TARY: ToleranceOperandType = None
    TARR: ToleranceOperandType = None
    ISOA: ToleranceOperandType = None
    ISOB: ToleranceOperandType = None
    ISOC: ToleranceOperandType = None
    ISOD: ToleranceOperandType = None
    MPVT: ToleranceOperandType = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = ("ColorRaysBy", "IAS_Spot", "Patterns", "Reference", "ShowScales")

class ColorRaysBy:
    Fields = 0
    Waves = 1
    Config = 2
    Wavelength = 3

class IAS_Spot(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def Pattern(self) -> Patterns: ...
    @Pattern.setter
    def Pattern(self, value: Patterns) -> None: ...
    @property
    def ReferTo(self) -> Reference: ...
    @ReferTo.setter
    def ReferTo(self, value: Reference) -> None: ...
    @property
    def ShowScale(self) -> ShowScales: ...
    @ShowScale.setter
    def ShowScale(self, value: ShowScales) -> None: ...
    @property
    def ColorRaysBy(self) -> ColorRaysBy: ...
    @ColorRaysBy.setter
    def ColorRaysBy(self, value: ColorRaysBy) -> None: ...
    @property
    def RayDensity(self) -> int: ...
    @RayDensity.setter
    def RayDensity(self, value: int) -> None: ...
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @property
    def DirectionCosines(self) -> bool: ...
    @DirectionCosines.setter
    def DirectionCosines(self, value: bool) -> None: ...
    @property
    def UseSymbols(self) -> bool: ...
    @UseSymbols.setter
    def UseSymbols(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def ScatterRays(self) -> bool: ...
    @ScatterRays.setter
    def ScatterRays(self, value: bool) -> None: ...
    @property
    def ShowAiryDisk(self) -> bool: ...
    @ShowAiryDisk.setter
    def ShowAiryDisk(self, value: bool) -> None: ...
    @property
    def IgnoreLateralColor(self) -> bool: ...
    @IgnoreLateralColor.setter
    def IgnoreLateralColor(self, value: bool) -> None: ...
    @property
    def PlotScale(self) -> float: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @property
    def DeltaFocus(self) -> float: ...
    @DeltaFocus.setter
    def DeltaFocus(self, value: float) -> None: ...
    @property
    def Exaggerate(self) -> float: ...
    @Exaggerate.setter
    def Exaggerate(self, value: float) -> None: ...

class Patterns:
    Square = 0
    Hexapolar = 1
    Dithered = 2

class Reference:
    ChiefRay = 0
    Centroid = 1
    Middle = 2
    Vertex = 3

class ShowScales:
    ScaleBar = 0
    Box = 1
    Cross = 2
    Circle = 3



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "CriterionComps",
    "CriterionFields",
    "Criterions",
    "MonteCarloStatistics",
    "QSCriterions",
    "SetupCaches",
    "SetupChanges",
    "SetupModes",
    "SetupPolynomials",
    "TolerancingColumnName",
    "TolerancingOperand",
    "TolerancingParameterName",
)

class CriterionComps:
    OptimizeAll_DLS: CriterionComps = None
    NONSEQOptimizeAll: CriterionComps = None
    ParaxialFocus: CriterionComps = None
    NONSEQNone: CriterionComps = None
    NONSEQOptimizeAll_OD: CriterionComps = None
    None_: CriterionComps = None
    OptimizeAll_OD: CriterionComps = None

class CriterionFields:
    NONSEQUserDefined: CriterionFields = None
    Y_Symmetric: CriterionFields = None
    XY_Symmetric: CriterionFields = None
    UserDefined: CriterionFields = None

class Criterions:
    NONSEQMeritFunction: Criterions = None
    RMSSpotRadius: Criterions = None
    RMSSpotX: Criterions = None
    NONSEQUserScript: Criterions = None
    RMSSpotY: Criterions = None
    RMSWavefront: Criterions = None
    MeritFunction: Criterions = None
    GeometricMTFAverage: Criterions = None
    GeometricMTFTan: Criterions = None
    GeometricMTFSag: Criterions = None
    DiffMTFAverage: Criterions = None
    DiffMTFTan: Criterions = None
    DiffMTFSag: Criterions = None
    BoresightError: Criterions = None
    RMSAngularRadius: Criterions = None
    RMSAngularX: Criterions = None
    RMSAngularY: Criterions = None
    UserScript: Criterions = None

class MonteCarloStatistics:
    Normal: MonteCarloStatistics = None
    Uniform: MonteCarloStatistics = None
    Parabolic: MonteCarloStatistics = None

class QSCriterions:
    RMSSpotRadius: QSCriterions = None
    RMSSpotX: QSCriterions = None
    RMSSpotY: QSCriterions = None
    RMSWavefront: QSCriterions = None
    BoresightError: QSCriterions = None
    RMSAngularRadius: QSCriterions = None
    RMSAngularX: QSCriterions = None
    RMSAngularY: QSCriterions = None

class SetupCaches:
    RecomputeAll: SetupCaches = None
    RecomputeChanged: SetupCaches = None
    UsePolynomial: SetupCaches = None

class SetupChanges:
    LinearDifference: SetupChanges = None
    RSSDifference: SetupChanges = None

class SetupModes:
    Sensitivity: SetupModes = None
    InverseLimit: SetupModes = None
    InverseIncrement: SetupModes = None
    SkipSensitivity: SetupModes = None

class SetupPolynomials:
    None_: SetupPolynomials = None
    ThreeMinusTerm: SetupPolynomials = None
    FiveMinusTerm: SetupPolynomials = None

class TolerancingColumnName:
    UserMeritFunction: TolerancingColumnName = None
    UserScript: TolerancingColumnName = None
    RmsSpotRadius: TolerancingColumnName = None
    RmsSpotX: TolerancingColumnName = None
    RmsSpotY: TolerancingColumnName = None
    RmsWavefrontError: TolerancingColumnName = None
    MtfGeometricAverage: TolerancingColumnName = None
    MtfGeometricTangential: TolerancingColumnName = None
    MtfGeometricSaggital: TolerancingColumnName = None
    MtfDiffractionAverage: TolerancingColumnName = None
    MtfDiffractionTangential: TolerancingColumnName = None
    MtfDiffractionSagittal: TolerancingColumnName = None
    BoresightError: TolerancingColumnName = None
    RmsAngularRadius: TolerancingColumnName = None
    RmsAngularX: TolerancingColumnName = None
    RmsAngularY: TolerancingColumnName = None
    BackFocusChange: TolerancingColumnName = None
    Unknown: TolerancingColumnName = None

class TolerancingOperand:
    NULL: TolerancingOperand = None
    TRAD: TolerancingOperand = None
    TCUR: TolerancingOperand = None
    TFRN: TolerancingOperand = None
    TTHI: TolerancingOperand = None
    TCON: TolerancingOperand = None
    TSDI: TolerancingOperand = None
    TSDX: TolerancingOperand = None
    TSDY: TolerancingOperand = None
    TSTX: TolerancingOperand = None
    TSTY: TolerancingOperand = None
    TIRX: TolerancingOperand = None
    TIRY: TolerancingOperand = None
    TIRR: TolerancingOperand = None
    TPAR: TolerancingOperand = None
    TEDV: TolerancingOperand = None
    TEDX: TolerancingOperand = None
    TEDY: TolerancingOperand = None
    TETX: TolerancingOperand = None
    TETY: TolerancingOperand = None
    TETZ: TolerancingOperand = None
    TUDX: TolerancingOperand = None
    TUDY: TolerancingOperand = None
    TUTX: TolerancingOperand = None
    TUTY: TolerancingOperand = None
    TUTZ: TolerancingOperand = None
    TIND: TolerancingOperand = None
    TABB: TolerancingOperand = None
    TEXI: TolerancingOperand = None
    SAVE: TolerancingOperand = None
    COMP: TolerancingOperand = None
    CPAR: TolerancingOperand = None
    CEDV: TolerancingOperand = None
    CMCO: TolerancingOperand = None
    STAT: TolerancingOperand = None
    TWAV: TolerancingOperand = None
    TNPS: TolerancingOperand = None
    TNPA: TolerancingOperand = None
    TMCO: TolerancingOperand = None
    TEZI: TolerancingOperand = None
    TCMU: TolerancingOperand = None
    SEED: TolerancingOperand = None
    COMM: TolerancingOperand = None
    TPAI: TolerancingOperand = None
    TCIO: TolerancingOperand = None
    TCEO: TolerancingOperand = None
    TNMA: TolerancingOperand = None
    CNPA: TolerancingOperand = None
    CNPS: TolerancingOperand = None
    TRLX: TolerancingOperand = None
    TRLY: TolerancingOperand = None
    TRLR: TolerancingOperand = None
    TARX: TolerancingOperand = None
    TARY: TolerancingOperand = None
    TARR: TolerancingOperand = None
    TEDR: TolerancingOperand = None
    TSDR: TolerancingOperand = None
    ISOA: TolerancingOperand = None
    ISOB: TolerancingOperand = None
    ISOC: TolerancingOperand = None
    ISOD: TolerancingOperand = None
    MPVT: TolerancingOperand = None
    Unknown: TolerancingOperand = None

class TolerancingParameterName:
    CompensatorSetting: TolerancingParameterName = None
    ConfigurationNumber: TolerancingParameterName = None
    CommentField: TolerancingParameterName = None
    NominalValue: TolerancingParameterName = None
    Surface: TolerancingParameterName = None
    MinimumValue: TolerancingParameterName = None
    MaximumValue: TolerancingParameterName = None
    Surface1: TolerancingParameterName = None
    Surface2: TolerancingParameterName = None
    AdjustSurface: TolerancingParameterName = None
    ParameterNumber: TolerancingParameterName = None
    Layer: TolerancingParameterName = None
    MaximumNumber: TolerancingParameterName = None
    MinimumNumber: TolerancingParameterName = None
    Code: TolerancingParameterName = None
    RowNumber: TolerancingParameterName = None
    ConfigurationParameter: TolerancingParameterName = None
    Object: TolerancingParameterName = None
    Data: TolerancingParameterName = None
    FieldNumber: TolerancingParameterName = None
    SamplingSetting: TolerancingParameterName = None
    Units: TolerancingParameterName = None
    Statistics: TolerancingParameterName = None
    PivotPointOption: TolerancingParameterName = None
    FieldSetting: TolerancingParameterName = None
    TdeRowNumber: TolerancingParameterName = None
    Undefined: TolerancingParameterName = None



"""Decorators for Pydantic dataclasses with default configurations."""

from __future__ import annotations

import dataclasses
from sys import version_info

import pydantic
from pydantic import ConfigDict, Field, PrivateAttr

if version_info <= (3, 11):
    from typing_extensions import dataclass_transform
else:
    from typing import dataclass_transform


__all__ = ("analysis_result", "analysis_settings")


def _default_config_dataclass(default_config: ConfigDict, cls=None, config: ConfigDict | None = None, **kwargs):
    """Pydantic dataclass with default configuration."""
    config = default_config if config is None else default_config.update(config)

    if cls is None:
        # Called with parentheses
        return pydantic.dataclasses.dataclass(config=config, **kwargs)

    return pydantic.dataclasses.dataclass(config=config, **kwargs)(cls)


@dataclass_transform(field_specifiers=(dataclasses.field, Field, PrivateAttr))
def analysis_result(cls=None, config: ConfigDict | None = None, **kwargs):
    """Pydantic dataclass with default configuration for analysis results."""
    default_config = ConfigDict(populate_by_name=True, ser_json_inf_nan="constants")
    return _default_config_dataclass(default_config, cls, config, **kwargs)


@dataclass_transform(field_specifiers=(dataclasses.field, Field, PrivateAttr))
def analysis_settings(cls=None, config: ConfigDict | None = None, **kwargs):
    """Pydantic dataclass with default configuration for analysis settings."""
    default_config = ConfigDict(validate_assignment=True)
    return _default_config_dataclass(default_config, cls, config, **kwargs)



"""Data types for analysis result and settings objects."""

from __future__ import annotations

import logging
from operator import attrgetter
from typing import TYPE_CHECKING, Annotated, Any, Generic, Literal, Optional, TypeVar, Union

from numpy import array, ndarray
from pandas import DataFrame
from pydantic import Field
from pydantic.dataclasses import dataclass
from pydantic_core import CoreSchema, PydanticCustomError, core_schema

from zospy.api import constants

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler

__all__ = ("UnitField", "ValidatedDataFrame", "ValidatedNDArray", "WavelengthNumber", "FieldNumber", "ZOSAPIConstant")


Value = TypeVar("Value")


@dataclass
class UnitField(Generic[Value]):
    """A field with a unit."""

    value: Value
    unit: str


class ValidatedDataFrameAnnotation:
    """Pydantic validation and serialization for Pandas DataFrames."""

    @staticmethod
    def _validate_dataframe(value: dict | DataFrame) -> DataFrame:
        if isinstance(value, dict):
            try:
                return DataFrame.from_dict(value, orient="tight")
            # Pandas can raise a KeyError if the dictionary was created with a different `orient`
            except (KeyError, ValueError) as e:
                raise PydanticCustomError(
                    "invalid_dataframe",
                    "Cannot convert dictionary to DataFrame: {value}",
                    {"value": value},
                ) from e

        return value

    @staticmethod
    def _serialize_dataframe(value: DataFrame) -> dict:
        return value.to_dict(orient="tight")

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        schema = core_schema.json_or_python_schema(
            json_schema=core_schema.dict_schema(),
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(DataFrame), core_schema.dict_schema()]
            ),
        )

        serializer = core_schema.plain_serializer_function_ser_schema(
            cls._serialize_dataframe, when_used="json-unless-none"
        )

        return core_schema.no_info_after_validator_function(
            cls._validate_dataframe,
            schema,
            serialization=serializer,
        )


ValidatedDataFrame = Annotated[DataFrame, ValidatedDataFrameAnnotation]


class ValidatedNDArrayAnnotation:
    """Pydantic validation and serialization for Numpy arrays."""

    @staticmethod
    def _serialize_ndarray(value: ndarray) -> list:
        return value.tolist()

    @staticmethod
    def _validate_ndarray(value: list | ndarray) -> ndarray:
        if isinstance(value, list):
            try:
                return array(value)
            except ValueError as e:
                raise PydanticCustomError(
                    "invalid_ndarray",
                    "Cannot convert list to ndarray: {value}",
                    {"value": value},
                ) from e

        return value

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        schema = core_schema.json_or_python_schema(
            json_schema=core_schema.list_schema(),
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(ndarray), core_schema.list_schema()]
            ),
        )

        serializer = core_schema.plain_serializer_function_ser_schema(
            cls._serialize_ndarray, when_used="json-unless-none"
        )

        return core_schema.no_info_after_validator_function(
            cls._validate_ndarray,
            schema,
            serialization=serializer,
        )


ValidatedNDArray = Annotated[ndarray, ValidatedNDArrayAnnotation]

ZospyConstantType = TypeVar("ZospyConstantType")


class ZOSAPIConstantAnnotation:
    """Pydantic validation and serialization for ZOSAPI constants."""

    def __init__(self, enum: str):
        self.enum = enum

    def _validate_constant(self, value: ZospyConstantType | str) -> ZospyConstantType:
        try:
            constant = attrgetter(self.enum)(constants)
        except AttributeError:
            logging.warning(f"Constant {self.enum} not found in zospy.constants")
            return None

        return constants.process_constant(constant, value)

    def __get_pydantic_core_schema__(self, source_type: Any, handler: GetCoreSchemaHandler) -> any:
        """Validate ZOSAPI constants."""
        schema = core_schema.json_or_python_schema(
            json_schema=core_schema.any_schema(),
            python_schema=core_schema.any_schema(),
        )

        serializer = core_schema.plain_serializer_function_ser_schema(str, when_used="json-unless-none")

        return core_schema.no_info_after_validator_function(self._validate_constant, schema, serialization=serializer)


def ZOSAPIConstant(enum: str) -> type[str]:  # noqa: N802
    """Pydantic validation and serialization for ZOSAPI constants."""
    return Annotated[Optional[str], ZOSAPIConstantAnnotation(enum)]


WavelengthNumber = Union[Literal["All"], Annotated[int, Field(gt=0)]]
FieldNumber = Union[Literal["All"], Annotated[int, Field(gt=0)]]



"""Zemax OpticStudio analyses from the Physical Optics category."""

from __future__ import annotations

import warnings
from typing import Literal

import pandas as pd

from zospy import utils
from zospy.analyses.old.base import Analysis, AnalysisResult, OnComplete, new_analysis
from zospy.api import constants
from zospy.zpcore import OpticStudioSystem


def pop_create_beam_parameter_dict(
    oss: OpticStudioSystem, beam_type: constants.Analysis.PhysicalOptics.POPBeamTypes | str = "GaussianWaist"
) -> dict:
    """Creates a dictionary containing the parameters for a certain Physical Optics beam type and their default values.

    The dictionary can be adjusted and supplied to the beam_parameters argument of physical_optics_propagation()

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    beam_type : constants.Analysis.PhysicalOptics.POPBeamTypes | str
        The beam type. Defaults to 'GaussianWaist'.

    Returns
    -------
    dict
        A dictionary containing the parameters for a certain Physical Optics beam type and their default values.
    """
    analysis_type = constants.Analysis.AnalysisIDM.PhysicalOpticsPropagation

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    analysis.Settings.BeamType = constants.process_constant(constants.Analysis.PhysicalOptics.POPBeamTypes, beam_type)

    beam_params = {}
    for i in range(analysis.Settings.NumberOfParameters):
        beam_params[analysis.Settings.GetParameterName(i)] = analysis.Settings.GetParameterValue(i)

    analysis.Close()

    return beam_params


def pop_create_fiber_parameter_dict(
    oss: OpticStudioSystem, fiber_type: constants.Analysis.PhysicalOptics.POPFiberTypes | str = "GaussianWaist"
) -> dict:
    """Creates a dictionary containing the parameters for a certain Physical Optics fiber type and their default values.

     The dictionary can be adjusted and supplied to the fiber_parameters argument of physical_optics_propagation()

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    fiber_type : constants.Analysis.PhysicalOptics.POPFiberTypes | str
         The fiber type. Defaults to 'GaussianWaist'.

    Returns
    -------
    dict
        A dictionary containing the parameters for a certain Physical Optics fiber type and their default values.
    """
    analysis_type = constants.Analysis.AnalysisIDM.PhysicalOpticsPropagation

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    analysis.Settings.FiberType = constants.process_constant(
        constants.Analysis.PhysicalOptics.POPFiberTypes, fiber_type
    )

    fiber_params = {}
    for i in range(analysis.Settings.NumberOfFiberParameters):
        fiber_params[analysis.Settings.GetFiberParameterName(i)] = analysis.Settings.GetFiberParameterValue(i)

    analysis.Close()

    return fiber_params


def _set_pop_parameters(analysis: Analysis, which: Literal["beam", "fiber"], parameters: dict) -> None:
    """Sets beam or fiber parameters using the provided dictionary.

    Only parameters present in the dictionary are set. After setting, the function checks if all provided parameters
    were used and raises an error if any are left unused.

    Parameters
    ----------
    analysis : Analysis
        The analysis object containing settings for either beam or fiber parameters.
    which : Literal['beam', 'fiber']
        Specifies whether to set 'beam' or 'fiber' parameters. Determines which set of methods to use within the
        `analysis` object.
    parameters : dict | None
        A dictionary containing parameter names and their desired values. See also pop_create_beam_parameter_dict() and
        pop_create_fiber_parameter_dict(). Parameters not accepted by the current beam or fiber type will raise an
        error.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If `which` is not 'beam' or 'fiber', or if any parameters in `parameters` do not match those expected for the
        specified `which` type in `analysis`.
    """
    if parameters is not None:
        parameters = parameters.copy()  # Create copy to preserve the original dictionary

        if which == "beam":
            number_of_parameters = analysis.Settings.NumberOfParameters
            get_parameter_name = analysis.Settings.GetParameterName
            set_parameter_value = analysis.Settings.SetParameterValue
        elif which == "fiber":
            number_of_parameters = analysis.Settings.NumberOfFiberParameters
            get_parameter_name = analysis.Settings.GetFiberParameterName
            set_parameter_value = analysis.Settings.SetFiberParameterValue
        else:
            raise ValueError("Invalid parameter type. Choose 'fiber' or 'beam'.")

        for i in range(number_of_parameters):
            param_name = get_parameter_name(i)
            param_value = parameters.pop(param_name, None)
            if param_value is not None:  # Only set parameters that are present in the dictionary
                set_parameter_value(i, param_value)

        if parameters:
            raise ValueError(
                f"The following {which} parameters are specified but not accepted: {parameters.keys()}. "
                f"The accepted {which} parameters for this {which}_type are: "
                f"{','.join([get_parameter_name(i) for i in range(number_of_parameters)])}"
            )


def physical_optics_propagation(
    oss: OpticStudioSystem,
    # General
    wavelength: Literal["All"] | int = "All",
    field: int = 1,
    start_surface: Literal["Ent. Pupil"] | int = 1,
    end_surface: Literal["Image"] | int = "Image",
    surface_to_beam: float = 0.0,
    use_polarization: bool = False,
    separate_xy: bool = False,
    use_disk_storage: bool = False,
    # Beam definition
    beam_type: constants.Analysis.PhysicalOptics.POPBeamTypes | str = "GaussianWaist",
    beam_file: str = "",
    beam_parameters: dict | None = None,
    x_sampling: constants.Analysis.SampleSizes | str | int = 32,
    y_sampling: constants.Analysis.SampleSizes | str | int = 32,
    x_width: float = 1.0,
    y_width: float = 1.0,
    use_total_power: bool = False,
    total_power: float = 1.0,
    use_peak_irradiance: bool = True,
    peak_irradiance: float = 1.0,
    # Display
    show_as: constants.Analysis.PhysicalOptics.POPShowAsTypes | str = "FalseColor",
    data_type: constants.Analysis.PhysicalOptics.POPDataTypes | str = "Irradiance",
    project: constants.Analysis.PhysicalOptics.POPProjectionTypes | str = "AlongBeam",
    contour_format: str = "",
    plot_scale: float = 0,
    scale_type: constants.Analysis.PhysicalOptics.POPScaleTypes | str = "Linear",
    zoom_in: constants.Analysis.PhysicalOptics.POPZoomTypes | str = "NoZoom",
    row_or_column: int | Literal["Center"] = "Center",
    zero_phase_level: float = 0.001,
    save_output_beam: bool = False,
    output_beam_file: str = "",
    save_beam_at_all_surfaces: bool = False,
    compute_fiber_coupling_integral: bool = False,
    ignore_fibre_polarization: bool = True,
    fiber_type: constants.Analysis.PhysicalOptics.POPFiberTypes | str = "GaussianWaist",
    fiber_position: constants.Analysis.PhysicalOptics.POPFiberPositions | str = "ChiefRay",
    tilt_about_x: float = 0.0,
    tilt_about_y: float = 0.0,
    fiber_type_file: str = "",
    fiber_parameters: dict | None = None,
    auto_calculate_beam_sampling: bool = False,
    oncomplete: OnComplete | str = OnComplete.Close,
) -> AnalysisResult:
    """Physical Optics Propagation analysis.

    See the OpticStudio documentation for a more in depth description of most parameters.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    wavelength : str | int
        The wavelength to use in the analysis. Either 'All' or an integer specifying the wavelength number. Defaults to
        'All'.
    field : int
        The field number of the chief ray used to align the beam for the analysis. Should be an integer specifying the
        field number. Defaults to 1.
    start_surface : str | int
        The surface that is to be analyzed. Either 'Ent. Pupil' or an integer specifying the surface number. Defaults to
         1.
    end_surface : str | int
        The surface that is to be analyzed. Either 'Image' or an integer specifying the surface number. Defaults to
        'Image'.
    surface_to_beam : float
        Distance from starting surface to starting beam position. Defaults to 0.0.
    use_polarization : bool
        Use polarization. Defaults to False.
    separate_xy : bool
        Separate propagation in X Y direction. Defaults to False.
    use_disk_storage : bool
        Use disk storage. Defaults to False.
    beam_type : constants.Analysis.PhysicalOptics.POPBeamTypes | str
        The beam type. Defaults to 'GaussianWaist'.
    beam_file:
        The beam file used for the beam definition. Should include the .ZBF extension. The beam files have to be located
        in the OpticStudio <pop> folder. Only used when beam_type is set to 'File', 'DLL' or 'Multimode'. Defaults to
        ''.
    beam_parameters : dict | None
        The beam parameters. Either a dictionary containing the beam parameters settings or None. Only the parameters
        that are defined in the dictionary will be set, the other parameters will remain their default value. Note that
        the accepted beam parameters vary per beam type. The accepted parameters for a specific beam type and their
        default values can be obtained through zospy.analysis.physicaloptics.pop_create_beam_parameter_dict(), which can
        then be adjusted and parsed to beam_parameters. When set to None, the default beam parameters will be used.
        Defaults to None.
    x_sampling : constants.Analysis.SampleSizes | str | int
        The number of points used to sample the beam in the x direction. Either specified using
        constants.Analysis.SampleSizes, a string (e.g. '32x32') or integer (e.g. 32). Note that the string or integer
        should match one of the sample sizes in constants.Analysis.SampleSizes (32 or '32x32' for 'S_32x32' etc.).
        Defaults to 32.
    y_sampling : constants.Analysis.SampleSizes | str | int
        The number of points used to sample the beam in the y direction. Either specified using
        constants.Analysis.SampleSizes, a string (e.g. '32x32') or integer (e.g. 32). Note that the string or integer
        should match one of the sample sizes in constants.Analysis.SampleSizes (32 or '32x32' for 'S_32x32' etc.).
        Defaults to 32.
    x_width : float
        The original with of the region represented by the array in OpticStudio lens units. Defaults to 1.0.
    y_width : float
        The original with of the region represented by the array in OpticStudio lens units. Defaults to 1.0.
    use_total_power : bool
        Use total power. Note that either use_total_power or use_peak_irradiance has to be True. Defaults to False.
    total_power : float
        The total power. Only used when use_total_power is True. Defaults to 1.0.
    use_peak_irradiance : bool
        Use peak irradiance. Note that either use_total_power or use_peak_irradiance has to be True. Defaults to True.
    peak_irradiance : float
        The peak irradiance. Only used when use_peak_irradiance is True. Defaults to 1.0.
    show_as : constants.Analysis.PhysicalOptics.POPShowAsTypes | str
        Defines the default graph in OpticStudio. Defaults to 'FalseColor'.
    data_type : constants.Analysis.PhysicalOptics.POPDataTypes | str
        Defines the type of output data. Defaults to 'Irradiance'.
    project : constants.Analysis.PhysicalOptics.POPProjectionTypes | str
        Defines the perspective from which the beam is viewed. Defaults to 'AlongBeam'.
    contour_format : str
        The contour format string. An empty string means no formatting. Only used when show_as is set to 'Contour'.
        Defaults to ''.
    plot_scale : float
        The plot scale. Defaults to 0.
    scale_type : constants.Analysis.PhysicalOptics.POPScaleTypes | str
        The scale type for the plot. Defaults to 'Linear'.
    zoom_in : constants.Analysis.PhysicalOptics.POPZoomTypes | str
        The zoom factor for the plot. Defaults to 'NoZoom'.
    row_or_column : int | str
        An integer defining the row or column number used when show_as is either 'CrossX' or 'CrossY'. Also accepts
        'Center'. Defaults to 'Center'.
    zero_phase_level : float
        The lower limit on the relative irradiance. Data points below this threshold have a phase value of 0. Defaults
        to 0.001.
    save_output_beam : bool
        Defines whether the output beam should be saved. Defaults to False.
    output_beam_file : str
        The filename to which the output beam should be saved. Should NOT include the .ZBF extension. The file will be
        saved in the OpticStudio <pop> folder.
    save_beam_at_all_surfaces : bool
        Defines whether the beam should be saved at all surfaces. Only used when save_output_beam is set to True.
        Defaults to False.
    compute_fiber_coupling_integral : bool
        Defines whether the fibre coupling is computed. Defaults to False.
    ignore_fibre_polarization :
        Defines whether polarization is considered. Defaults to True.
    fiber_type : constants.Analysis.PhysicalOptics.POPFiberTypes | str
        The fiber type. Defaults to 'GaussianWaist'.
    fiber_position : constants.Analysis.PhysicalOptics.POPFiberPositions | str
        The reference for the center of the receiving fiber. Defaults t 'ChiefRay'.
    tilt_about_x :
        The tilt of the fiber around the x-axis in degrees with respect to the beam. Defaults to 0.0.
    tilt_about_y :
        The tilt of the fiber around the y-axis in degrees with respect to the beam. Defaults to 0.0.
    fiber_type_file :
        The file containing the fiber description. Only used is fiber_type is set to 'File' or 'DLL'.
    fiber_parameters : dict | None
        The fiber parameters. Either a dictionary containing the fiber parameters settings or None. Only the parameters
        that are defined in the dictionary will be set, the other parameters will remain their default value. Note that
        the accepted beam parameters vary per fiber type. The accepted parameters for a specific fiber type and their
        default values can be obtained through zospy.analysis.physicaloptics.pop_create_fiber_parameter_dict(), which
        can then be adjusted and parsed to fiber_parameters. When set to None, the default fiber parameters will be
        used. Defaults to None.
    auto_calculate_beam_sampling : bool
        Defines whether the beam sampling should be automatically calculated. Will adjust x_width and y_width. This
        calculation is done just before the analysis is run. Only used when beam_type is not one of ('File', 'DLL',
        'Multimode'). Defaults to False.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.

    Returns
    -------
    AnalysisResult
        A Physical Optics Propagation analysis result.
    """
    # TODO: evaluate what happens when output_beam_file includes either a filepath or the .ZBF extension.
    analysis_type = constants.Analysis.AnalysisIDM.PhysicalOpticsPropagation

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    # Adjust settings
    analysis.set_wavelength(wavelength)
    analysis.set_field(field)

    # Start surface
    if start_surface == "Ent. Pupil":
        analysis.Settings.StartSurface.SetSurfaceNumber(0)
    elif isinstance(start_surface, int):
        analysis.Settings.StartSurface.SetSurfaceNumber(start_surface)
    else:
        raise ValueError(f'start_surface value should be "Ent. Pupil" or an integer, got {start_surface}')

    # End surface
    if end_surface == "Image":
        analysis.Settings.EndSurface.UseImageSurface()
    elif isinstance(end_surface, int):
        analysis.Settings.EndSurface.SetSurfaceNumber(end_surface)
    else:
        raise ValueError(f'end_surface value should be "Image" or an integer, got {end_surface}')

    analysis.Settings.SurfaceToBeam = surface_to_beam
    analysis.Settings.UsePolarization = use_polarization
    analysis.Settings.SeparateXY = separate_xy
    analysis.Settings.UseDiskStorage = use_disk_storage

    # Beam settings
    analysis.Settings.BeamType = constants.process_constant(constants.Analysis.PhysicalOptics.POPBeamTypes, beam_type)
    if isinstance(x_sampling, int):
        x_sampling = f"{x_sampling}x{x_sampling}"
    analysis.Settings.XSampling = constants.process_constant(
        constants.Analysis.SampleSizes, utils.zputils.standardize_sampling(x_sampling)
    )
    if isinstance(y_sampling, int):
        y_sampling = f"{y_sampling}x{y_sampling}"
    analysis.Settings.YSampling = constants.process_constant(
        constants.Analysis.SampleSizes, utils.zputils.standardize_sampling(y_sampling)
    )

    analysis.Settings.XWidth = x_width
    analysis.Settings.YWidth = y_width

    if not (use_total_power ^ use_peak_irradiance):
        raise ValueError(
            "Either use_total_power or use_peak_irradiance should be True, they cannot both be True or False."
        )
    if use_total_power:
        analysis.Settings.UseTotalPower = use_total_power
        analysis.Settings.TotalPower = total_power
    else:
        analysis.Settings.UsePeakIrradiance = use_peak_irradiance
        analysis.Settings.PeakIrradiance = peak_irradiance

    if beam_file != "":
        if str(analysis.Settings.BeamType) in ("File", "DLL", "Multimode"):
            analysis.Settings.BeamTypeFilename = beam_file
        else:
            raise ValueError(f"Beam type {beam_type!s} does not allow specification of beam_file.")

    _set_pop_parameters(analysis, which="beam", parameters=beam_parameters)

    # Display settings
    analysis.Settings.ShowAs = constants.process_constant(constants.Analysis.PhysicalOptics.POPShowAsTypes, show_as)
    analysis.Settings.DataType = constants.process_constant(constants.Analysis.PhysicalOptics.POPDataTypes, data_type)
    analysis.Settings.Project = constants.process_constant(
        constants.Analysis.PhysicalOptics.POPProjectionTypes, project
    )

    if str(analysis.Settings.ShowAs) == "Contour":
        analysis.Settings.ContourFormat = contour_format
    elif str(analysis.Settings.ShowAs) in ("CrossX", "CrossY"):
        analysis.Settings.RowOrColumn = 0 if row_or_column == "Center" else row_or_column

    analysis.Settings.PlotScale = plot_scale
    analysis.Settings.ScaleType = constants.process_constant(
        constants.Analysis.PhysicalOptics.POPScaleTypes, scale_type
    )
    analysis.Settings.ZoomIn = constants.process_constant(constants.Analysis.PhysicalOptics.POPZoomTypes, zoom_in)
    analysis.Settings.ZeroPhaseLevel = zero_phase_level

    analysis.Settings.SaveOutputBeam = save_output_beam
    if analysis.Settings.SaveOutputBeam:
        if output_beam_file == "":
            raise ValueError("output_beam_file should be a valid filepath if save_output_beam is True.")
        analysis.Settings.OutputBeamFile = output_beam_file
    else:
        if output_beam_file != "":
            raise warnings.warn("save_output_beam is False but output_beam_file is specified.")

        analysis.Settings.SaveBeamAtAllSurfaces = save_beam_at_all_surfaces

    analysis.Settings.ComputeFiberCouplingIntegral = compute_fiber_coupling_integral

    # Fiber settings
    if analysis.Settings.ComputeFiberCouplingIntegral:
        analysis.Settings.IgnoreFiberPolarization = ignore_fibre_polarization
        analysis.Settings.FiberType = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPFiberTypes, fiber_type
        )
        analysis.Settings.FiberPosition = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPFiberPositions, fiber_position
        )
        analysis.Settings.TiltAboutX = tilt_about_x
        analysis.Settings.TiltAboutY = tilt_about_y

        if fiber_type_file != "":
            if str(analysis.Settings.FiberType) in ("File", "DLL"):
                analysis.Settings.FiberTypeFilename = fiber_type_file
            else:
                raise ValueError(
                    f"Beam type {analysis.Settings.FiberTypeFilename!s} does not allow specification of fiber_type_file."
                )

        _set_pop_parameters(analysis, which="fiber", parameters=fiber_parameters)

    # Auto calculate beam sampling
    if auto_calculate_beam_sampling and str(analysis.Settings.BeamType) not in ("File", "DLL", "Multimode"):
        analysis.Settings.AutoCalculateBeamSampling()

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["Field"] = analysis.get_field()
    if analysis.Settings.StartSurface.GetSurfaceNumber() == 0:
        settings.loc["StartSurface"] = "Ent. Pupil"
    else:
        settings.loc["StartSurface"] = analysis.Settings.StartSurface.GetSurfaceNumber()

    if analysis.Settings.EndSurface.GetSurfaceNumber() == 0:
        settings.loc["EndSurface"] = "Image"
    else:
        settings.loc["EndSurface"] = analysis.Settings.EndSurface.GetSurfaceNumber()

    settings.loc["SurfaceToBeam"] = analysis.Settings.SurfaceToBeam
    settings.loc["UsePolarization"] = analysis.Settings.UsePolarization
    settings.loc["SeparateXY"] = analysis.Settings.SeparateXY
    settings.loc["UseDiskStorage"] = analysis.Settings.UseDiskStorage
    settings.loc["BeamType"] = str(analysis.Settings.BeamType)
    settings.loc["XSampling"] = str(analysis.Settings.XSampling)
    settings.loc["YSampling"] = str(analysis.Settings.YSampling)
    settings.loc["XWidth"] = analysis.Settings.XWidth
    settings.loc["YWidth"] = analysis.Settings.YWidth
    settings.loc["UseTotalPower"] = analysis.Settings.UseTotalPower
    if analysis.Settings.UseTotalPower:
        settings.loc["TotalPower"] = analysis.Settings.TotalPower

    settings.loc["UsePeakIrradiance"] = analysis.Settings.UsePeakIrradiance
    if analysis.Settings.UsePeakIrradiance:
        settings.loc["PeakIrradiance"] = analysis.Settings.PeakIrradiance

    if str(analysis.Settings.BeamType) in ("File", "DLL", "Multimode"):
        settings.loc["BeamTypeFilename"] = analysis.Settings.BeamTypeFilename

    for i in range(analysis.Settings.NumberOfParameters):
        key = "Beam" + "".join(analysis.Settings.GetParameterName(i).split(" "))
        settings.loc[key] = analysis.Settings.GetParameterValue(i)

    settings.loc["ShowAs"] = str(analysis.Settings.ShowAs)
    settings.loc["DataType"] = str(analysis.Settings.DataType)
    settings.loc["Project"] = str(analysis.Settings.Project)

    if str(analysis.Settings.ShowAs) == "Contour":
        settings.loc["ContourFormat"] = analysis.Settings.ContourFormat
    elif str(analysis.Settings.ShowAs) in ("CrossX", "CrossY"):
        if analysis.Settings.RowOrColumn == 0:
            settings.loc["RowOrColumn"] = "Center"
        else:
            settings.loc["RowOrColumn"] = analysis.Settings.RowOrColumn

    settings.loc["PlotScale"] = analysis.Settings.PlotScale
    settings.loc["ScaleType"] = str(analysis.Settings.LogScale)
    settings.loc["ZoomIn"] = str(analysis.Settings.ZoomIn)
    settings.loc["ZeroPhaseLevel"] = analysis.Settings.ZeroPhaseLevel

    settings.loc["SaveOutputBeam"] = analysis.Settings.SaveOutputBeam
    if analysis.Settings.SaveOutputBeam:
        settings.loc["OutputBeamFile"] = analysis.Settings.OutputBeamFile
        settings.loc["SaveBeamAtAllSurfaces"] = analysis.Settings.SaveBeamAtAllSurfaces

    settings.loc["ComputeFiberCouplingIntegral"] = analysis.Settings.ComputeFiberCouplingIntegral

    if analysis.Settings.ComputeFiberCouplingIntegral:
        settings.loc["ComputeFiberCouplingIntegral"] = analysis.Settings.ComputeFiberCouplingIntegral
        settings.loc["IgnoreFiberPolarization"] = analysis.Settings.IgnoreFiberPolarization
        settings.loc["FiberType"] = str(analysis.Settings.FiberType)
        settings.loc["FiberPosition"] = str(analysis.Settings.FiberPosition)
        settings.loc["TiltAboutX"] = analysis.Settings.TiltAboutX
        settings.loc["TiltAboutY"] = analysis.Settings.TiltAboutY

        if str(analysis.Settings.FiberType) in ("File", "DLL"):
            settings.loc["FiberTypeFilename"] = analysis.Settings.FiberTypeFilename

        for i in range(analysis.Settings.NumberOfFiberParameters):
            key = "Fiber" + "".join(analysis.Settings.GetFiberParameterName(i).split(" "))
            settings.loc[key] = analysis.Settings.GetFiberParameterValue(i)

    if str(analysis.Settings.BeamType) not in ("File", "DLL", "Multimode"):
        settings.loc["AutoCalculateBeamSampling"] = auto_calculate_beam_sampling

    # Get data and unpack
    data = []
    for i in range(analysis.Results.NumberOfDataGrids):
        data.append(utils.zputils.unpack_datagrid(analysis.Results.DataGrids[i],
                                                  # pass "center" and None for consistency with older ZOSPy versions
                                                  cell_origin="center", label_rounding=None))

    if len(data) == 0:
        data = pd.DataFrame()
    elif len(data) == 1:
        data = data[0]
    else:
        data = pd.concat(data, axis=1)

    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)



from __future__ import annotations

from dataclasses import fields, is_dataclass

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal


def assert_dataclass_equal(
    a, b, *, return_bool: bool = False, tolerance: float = 1e-5, ignore_fields: list[str] | None = None
) -> None | bool:
    if not is_dataclass(a) and not is_dataclass(b):
        raise ValueError("Both result and expected must be dataclasses")

    fields_a, fields_b = fields(a), fields(b)

    missing_fields = {f.name for f in fields_a}.symmetric_difference(f.name for f in fields_b)
    differences = {}

    ignore_fields = ignore_fields or []

    for f in fields(a):
        if f.name in missing_fields or f.name in ignore_fields:
            continue

        value_a = getattr(a, f.name)
        value_b = getattr(b, f.name)

        if isinstance(value_a, pd.DataFrame) and isinstance(value_b, pd.DataFrame):
            try:
                assert_frame_equal(value_a, value_b, rtol=tolerance)
            except AssertionError:
                differences[f.name] = (value_a, value_b)
        elif isinstance(value_a, np.ndarray) and isinstance(value_b, np.ndarray):
            if not np.allclose(value_a, value_b, rtol=tolerance):
                differences[f.name] = (value_a, value_b)
        elif is_dataclass(value_a) and is_dataclass(value_b):
            if not assert_dataclass_equal(value_a, value_b, return_bool=True):
                differences[f.name] = (value_a, value_b)
        elif isinstance(value_a, (list, tuple)) and isinstance(value_b, (list, tuple)):
            if is_dataclass(value_a[0]):
                for i, (item_a, item_b) in enumerate(zip(value_a, value_b)):
                    if not assert_dataclass_equal(item_a, item_b, return_bool=True):
                        differences[f"{f.name}[{i}]"] = (item_a, item_b)
        elif value_a != value_b:
            differences[f.name] = (value_a, value_b)

    if missing_fields or differences:
        if return_bool:
            return False

        missing_fields_message = (
            f"Fields: {''.join(missing_fields)} are not present in both dataclasses." if (missing_fields) else ""
        )
        differences_message = f"The following fields are unequal: {differences}" if differences else ""
        message = (
            f"{missing_fields_message} {differences_message}"
            if missing_fields_message and differences_message
            else (missing_fields_message or differences_message)
        )

        raise AssertionError(message)

    if return_bool:
        return True

    return None



"""Utility functions for the Non-Sequential Component Editor (NCE) in OpticStudio."""

from __future__ import annotations

from zospy.api import _ZOSAPI, constants

__all__ = (
    "find_object_by_comment",
    "object_change_type",
)


def object_change_type(obj: _ZOSAPI.Editors.NCE.INCERow, new_type: constants.Editors.NCE.ObjectType | str):
    """Change the object type in the Non-Sequential Component Editor.

    Parameters
    ----------
    obj : ZOSAPI.Editors.NCE.INCERow
        The Row/Object for which the change is to be made.
    new_type : zospy.constants.Editors.NCE.ObjectType | str
        The new object type, either string (e.g. 'StandardLens') or int. The integer will be treated as if obtained from
        zp.constants.Editors.NCE.ObjectType.

    Returns
    -------
    None

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> oss = zos.connect()
    >>> oss.make_nonsequential()
    >>> new_object = oss.NCE.InsertNewObjectAt(1)
    >>> zp.functions.nce.object_change_type(
    ...     new_object, zp.constants.Editors.NCE.ObjectType.StandardLens
    ... )
    """
    new_type = constants.process_constant(constants.Editors.NCE.ObjectType, new_type)

    # Apply
    new_surface_type_settings = obj.GetObjectTypeSettings(new_type)
    obj.ChangeType(new_surface_type_settings)


def find_object_by_comment(
    nce: _ZOSAPI.Editors.NCE, comment: str, *, case_sensitive: bool = True
) -> list[_ZOSAPI.Editors.NCE.INCERow]:
    """Retrieve objects from the Non-Sequential Component Editor that have the supplied string as Comment.

    In case of multiple matches, the objects are returned in ascending order.

    Parameters
    ----------
    nce : ZOSAPI.Editors.NCE
        The Non-sequential Component Editor (NCE).
    comment : str
        String that is searched for in the Comment column of the NCE.
    case_sensitive : bool=False
        Flag that specifies whether the search is case-sensitive (default value) or not.

    Returns
    -------
    list[ZOSAPI.Editors.NCE.INCERow]
        A list of object in the NCE that have a Comment column value which matches the comment argument.

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> oss = zos.connect()
    >>> oss.make_nonsequential()
    >>> nce_object_1 = oss.NCE.InsertNewObjectAt(1)
    >>> nce_object_1.Comment = "aa"
    >>> nce_object_2 = oss.NCE.InsertNewObjectAt(1)
    >>> nce_object_2.Comment = "bb"
    >>> nce_object_3 = oss.NCE.InsertNewObjectAt(1)
    >>> nce_object_3.Comment = "aA"
    >>> find_object_by_comment(oss.NCE, "aa")
    """
    # Number of objects in the NCE
    number_of_objects = nce.NumberOfObjects

    # Is the search case-sensitive?
    if not case_sensitive:
        # If the search is NOT case-sensitive put comment argument in all small letters
        comment = comment.lower()

    # Initialize list of objects with matching comment
    matching_objects = []

    # Loop over the objects in the NCE and check if their comment matches
    for object_index in range(1, number_of_objects + 1):
        # Retrieve current object
        current_object = nce.GetObjectAt(object_index)

        # Retrieve current object comment
        object_comment = current_object.Comment

        # Is the search case-sensitive?
        if not case_sensitive:
            # If the search is NOT case-sensitive put current object comment in all small letters
            object_comment = object_comment.lower()

        # If the comment matches, append the corresponding object to the return list
        if comment == object_comment:
            matching_objects.append(current_object)

    # Return list of objects with matching comment
    return matching_objects



"""Cardinal points analysis."""

from __future__ import annotations

from typing import Annotated, Literal

from pydantic import Field, model_validator
from pydantic_core._pydantic_core import PydanticCustomError

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers import ZospyTransformer
from zospy.analyses.parsers.transformers import SimpleField

__all__ = ("CardinalPoints", "CardinalPointsSettings")


class CardinalPointsTransformer(ZospyTransformer):
    def cardinal_points(self, args):
        header, *values = args

        updated_values = [SimpleField(v.name, dict(zip(header, v.value))) for v in values]

        return SimpleField("Cardinal Points", self.dict(updated_values))

    def cardinal_point(self, args):
        name, *values = args

        return SimpleField(str(name), values)


@analysis_result
class CardinalPoint:
    object: float = Field(alias="Object Space")
    image: float = Field(alias="Image Space")


@analysis_result
class CardinalPointSpecification:
    focal_length: CardinalPoint = Field(alias="Focal Length")
    focal_planes: CardinalPoint = Field(alias="Focal Planes")
    principal_planes: CardinalPoint = Field(alias="Principal Planes")
    anti_principal_planes: CardinalPoint = Field(alias="Anti-Principal Planes")
    nodal_planes: CardinalPoint = Field(alias="Nodal Planes")
    anti_nodal_planes: CardinalPoint = Field(alias="Anti-Nodal Planes")


@analysis_result
class CardinalPointsResult:
    starting_surface: int = Field(alias="Starting surface")
    ending_surface: int = Field(alias="Ending surface")
    wavelength: float = Field(alias="Wavelength")
    orientation: str = Field(alias="Orientation")
    lens_units: str = Field(alias="Lens units")
    cardinal_points: CardinalPointSpecification = Field(alias="Cardinal Points")


@analysis_settings
class CardinalPointsSettings:
    """Settings for the Cardinal Points analysis.

    Attributes
    ----------
    surface_1 : int
        The surface number corresponding to the first surface of the analyzed system. Defaults to 1.
    surface_2 : int
        The surface number corresponding to the last surface of the analyzed system, or "Image". Defaults to "Image".
    wavelength : int
        The wavelength number. Defaults to 1.
    orientation : Literal["Y-Z", "X-Z"]
        The orientation along which the cardinal points are calculated. Must be one of "X-Z", "Y-Z". Defaults to "Y-Z".
    """

    surface_1: Annotated[int, Field(ge=1)] = Field(default=1, description="First surface of the analyzed system")
    surface_2: Annotated[int, Field(ge=2)] | Literal["Image"] = Field(
        default="Image", description="Last surface of the analyzed system"
    )
    wavelength: int = Field(default=1, ge=1, description="Wavelength number")
    orientation: Literal["Y-Z", "X-Z"] = Field(
        default="Y-Z", description="Orientation along which the cardinal points are calculated"
    )

    @model_validator(mode="after")
    def validate_surfaces(self):
        """Validate that surface_1 is less than surface_2."""
        if isinstance(self.surface_1, int) and isinstance(self.surface_2, int) and self.surface_1 >= self.surface_2:
            raise PydanticCustomError(
                "invalid_surface_number",
                "surface_1 must be less than surface_2, but got {surface_1} >= {surface_2}",
                {"surface_1": self.surface_1, "surface_2": self.surface_2},
            )

        return self


class CardinalPoints(
    BaseAnalysisWrapper[CardinalPointsResult, CardinalPointsSettings],
    analysis_type="CardinalPoints",
    needs_text_output_file=True,
    needs_config_file=True,
):
    """Cardinal points analysis."""

    def __init__(
        self,
        *,
        surface_1: int = 1,
        surface_2: int | Literal["Image"] = "Image",
        wavelength: int = 1,
        orientation: Literal["Y-Z", "X-Z"] = "Y-Z",
    ):
        """Create a new Cardinal Points analysis.

        See Also
        --------
        CardinalPointsSettings : Settings for the Cardinal Points analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> CardinalPointsResult:
        """Run the Cardinal Points analysis."""
        if self.settings.surface_2 == "Image":
            self.settings.surface_2 = self.oss.LDE.NumberOfSurfaces - 1

        settings = self.analysis.GetSettings()
        settings.SaveTo(str(self.config_file))

        settings_bytestring = self.config_file.read_bytes()
        settings_bytearray = bytearray(settings_bytestring)
        settings_bytearray[20] = self.settings.surface_1  # byte 20 maps to the first surface
        settings_bytearray[24] = self.settings.surface_2  # byte 24 maps to the last surface
        settings_bytearray[28] = self.settings.wavelength  # byte 28 maps to the wavelength
        settings_bytearray[32] = 0 if self.settings.orientation == "Y-Z" else 1  # byte 32 maps to the orientation

        self.config_file.write_bytes(settings_bytearray)
        settings.LoadFrom(str(self.config_file))

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Read text file and parse to object
        return self.parse_output(
            "cardinal_points", transformer=CardinalPointsTransformer, result_type=CardinalPointsResult
        )



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = ("IAS_RTA_Angle", "IAS_RTA_Wavelength")

class IAS_RTA_Angle(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def ObjectNumber(self) -> int: ...
    @ObjectNumber.setter
    def ObjectNumber(self, value: int) -> None: ...
    @property
    def FaceNumber(self) -> int: ...
    @FaceNumber.setter
    def FaceNumber(self, value: int) -> None: ...
    @property
    def MinAngle(self) -> float: ...
    @MinAngle.setter
    def MinAngle(self, value: float) -> None: ...
    @property
    def MaxAngle(self) -> float: ...
    @MaxAngle.setter
    def MaxAngle(self, value: float) -> None: ...
    @property
    def MinY(self) -> float: ...
    @MinY.setter
    def MinY(self, value: float) -> None: ...
    @property
    def MaxY(self) -> float: ...
    @MaxY.setter
    def MaxY(self, value: float) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_RTA_Wavelength(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def ObjectNumber(self) -> int: ...
    @ObjectNumber.setter
    def ObjectNumber(self, value: int) -> None: ...
    @property
    def FaceNumber(self) -> int: ...
    @FaceNumber.setter
    def FaceNumber(self, value: int) -> None: ...
    @property
    def MinWavelength(self) -> float: ...
    @MinWavelength.setter
    def MinWavelength(self, value: float) -> None: ...
    @property
    def MaxWavelength(self) -> float: ...
    @MaxWavelength.setter
    def MaxWavelength(self, value: float) -> None: ...
    @property
    def MinY(self) -> float: ...
    @MinY.setter
    def MinY(self, value: float) -> None: ...
    @property
    def MaxY(self) -> float: ...
    @MaxY.setter
    def MaxY(self, value: float) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "CriterionTypes",
    "DefaultAndDegrees",
    "DefaultAndFringes",
    "OptimizationTypes",
    "PupilArmsCount",
    "ReferenceTypes",
    "ToleranceGrade",
    "ToleranceVendor",
    "WizardType",
)

class CriterionTypes:
    Wavefront: CriterionTypes = None
    Contrast: CriterionTypes = None
    Spot: CriterionTypes = None
    Angular: CriterionTypes = None

class DefaultAndDegrees:
    Default: DefaultAndDegrees = None
    Degrees: DefaultAndDegrees = None

class DefaultAndFringes:
    Default: DefaultAndFringes = None
    Fringes: DefaultAndFringes = None
    Percent: DefaultAndFringes = None

class OptimizationTypes:
    RMS: OptimizationTypes = None
    PTV: OptimizationTypes = None

class PupilArmsCount:
    Arms_6: PupilArmsCount = None
    Arms_8: PupilArmsCount = None
    Arms_10: PupilArmsCount = None
    Arms_12: PupilArmsCount = None

class ReferenceTypes:
    Centroid: ReferenceTypes = None
    ChiefRay: ReferenceTypes = None
    Unreferenced: ReferenceTypes = None

class ToleranceGrade:
    Commercial: ToleranceGrade = None
    Precision: ToleranceGrade = None
    HighPrecision: ToleranceGrade = None
    CellPhoneLens: ToleranceGrade = None

class ToleranceVendor:
    Asphericon: ToleranceVendor = None
    EdmundOptics: ToleranceVendor = None
    Generic: ToleranceVendor = None
    LaCroix: ToleranceVendor = None
    Optimax: ToleranceVendor = None

class WizardType:
    NSCOptimization: WizardType = None
    NSCBitmap: WizardType = None
    NSCRoadwayLighting: WizardType = None
    SEQOptimization: WizardType = None
    NSCTolerance: WizardType = None
    SEQTolerance: WizardType = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "CriterionTypes",
    "DefaultAndDegrees",
    "DefaultAndFringes",
    "IImageData",
    "INSCBitmapWizard",
    "INSCOptimizationWizard",
    "INSCRoadwayLightingWizard",
    "INSCToleranceWizard",
    "INSCWizard",
    "ISEQOptimizationWizard",
    "ISEQOptimizationWizard2",
    "ISEQToleranceWizard",
    "IToleranceWizard",
    "IWizard",
    "OptimizationTypes",
    "PupilArmsCount",
    "ReferenceTypes",
    "ToleranceGrade",
    "ToleranceVendor",
    "WizardType",
)

class CriterionTypes:
    Wavefront = 0
    Contrast = 1
    Spot = 2
    Angular = 3

class DefaultAndDegrees:
    Default = 0
    Degrees = 1

class DefaultAndFringes:
    Default = 0
    Fringes = 1
    Percent = 2

class IImageData:
    @property
    def ImageName(self) -> str: ...
    @property
    def Width(self) -> int: ...
    @property
    def Height(self) -> int: ...
    @property
    def BitsPerPixel(self) -> int: ...
    @property
    def Channels(self) -> int: ...
    @property
    def Stride(self) -> int: ...
    @property
    def IsRGB(self) -> bool: ...
    def GetPixels(self) -> list[int]: ...
    def GetPixelsSafe(self, totalSize: int, height: int, width: int) -> tuple[list[int]]: ...
    def GetRawData(self) -> list[int]: ...
    def GetRawDataSafe(self, totalSize: int, height: int, width: int) -> tuple[list[int]]: ...

class INSCBitmapWizard(INSCWizard, IWizard):
    @property
    def NSCSettings(self) -> INSCWizard: ...
    @property
    def CommonSettings(self) -> IWizard: ...
    @property
    def NumberOfBitmapFiles(self) -> int: ...
    def GetBitmapFileAt(self, idx: int) -> str: ...
    def GetPreviewImage(self) -> IImageData: ...

class INSCOptimizationWizard(INSCWizard, IWizard):
    @property
    def NSCSettings(self) -> INSCWizard: ...
    @property
    def CommonSettings(self) -> IWizard: ...

class INSCRoadwayLightingWizard(IWizard):
    @property
    def CommonSettings(self) -> IWizard: ...
    @property
    def IsNSCRoadwayLightingWizard(self) -> bool: ...
    @property
    def IsSplitRaysUsed(self) -> bool: ...
    @IsSplitRaysUsed.setter
    def IsSplitRaysUsed(self, value: bool) -> None: ...
    @property
    def IsScatterRaysUsed(self) -> bool: ...
    @IsScatterRaysUsed.setter
    def IsScatterRaysUsed(self, value: bool) -> None: ...
    @property
    def IsUsePolarizationUsed(self) -> bool: ...
    @IsUsePolarizationUsed.setter
    def IsUsePolarizationUsed(self, value: bool) -> None: ...
    @property
    def IsIgnoreErrorsUsed(self) -> bool: ...
    @IsIgnoreErrorsUsed.setter
    def IsIgnoreErrorsUsed(self, value: bool) -> None: ...
    @property
    def Arrangement(self) -> int: ...
    @Arrangement.setter
    def Arrangement(self, value: int) -> None: ...
    @property
    def Origin(self) -> int: ...
    @Origin.setter
    def Origin(self, value: int) -> None: ...
    @property
    def NumberOfLanes(self) -> int: ...
    @NumberOfLanes.setter
    def NumberOfLanes(self, value: int) -> None: ...
    @property
    def SurfaceClassification(self) -> int: ...
    @SurfaceClassification.setter
    def SurfaceClassification(self, value: int) -> None: ...
    @property
    def RoadClass(self) -> int: ...
    @RoadClass.setter
    def RoadClass(self, value: int) -> None: ...
    @property
    def StartAt(self) -> int: ...
    @StartAt.setter
    def StartAt(self, value: int) -> None: ...
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @property
    def MountingHeight(self) -> float: ...
    @MountingHeight.setter
    def MountingHeight(self, value: float) -> None: ...
    @property
    def LongitudinalSpacing(self) -> float: ...
    @LongitudinalSpacing.setter
    def LongitudinalSpacing(self, value: float) -> None: ...
    @property
    def LateralOffset(self) -> float: ...
    @LateralOffset.setter
    def LateralOffset(self, value: float) -> None: ...
    @property
    def LaneWidth(self) -> float: ...
    @LaneWidth.setter
    def LaneWidth(self, value: float) -> None: ...
    @property
    def OverallWeight(self) -> float: ...
    @OverallWeight.setter
    def OverallWeight(self, value: float) -> None: ...
    @property
    def NumberOfOrigins(self) -> int: ...
    @property
    def NumberOfConfigurations(self) -> int: ...
    @property
    def NumberOfArrangements(self) -> int: ...
    @property
    def NumberOfClassifications(self) -> int: ...
    @property
    def NumberOfRoadClasses(self) -> int: ...
    def GetArrangementAt(self, idx: int) -> str: ...
    def GetClassificationAt(self, idx: int) -> str: ...
    def GetConfigurationAt(self, idx: int) -> str: ...
    def GetOriginAt(self, idx: int) -> str: ...
    def GetRoadClassAt(self, idx: int) -> str: ...

class INSCToleranceWizard(IToleranceWizard, IWizard):
    @property
    def ToleranceSettings(self) -> IToleranceWizard: ...

class INSCWizard(IWizard):
    @property
    def IsNSCOptimizingWizard(self) -> bool: ...
    @property
    def IsNSCBitmapWizard(self) -> bool: ...
    @property
    def IsClearDataSettingsUsed(self) -> bool: ...
    @IsClearDataSettingsUsed.setter
    def IsClearDataSettingsUsed(self, value: bool) -> None: ...
    @property
    def IsRaytraceSettingsUsed(self) -> bool: ...
    @IsRaytraceSettingsUsed.setter
    def IsRaytraceSettingsUsed(self, value: bool) -> None: ...
    @property
    def IsSplitRaysUsed(self) -> bool: ...
    @IsSplitRaysUsed.setter
    def IsSplitRaysUsed(self, value: bool) -> None: ...
    @property
    def IsUsePolarizationUsed(self) -> bool: ...
    @IsUsePolarizationUsed.setter
    def IsUsePolarizationUsed(self, value: bool) -> None: ...
    @property
    def IsScatterRaysUsed(self) -> bool: ...
    @IsScatterRaysUsed.setter
    def IsScatterRaysUsed(self, value: bool) -> None: ...
    @property
    def IsIgnoreErrorsUsed(self) -> bool: ...
    @IsIgnoreErrorsUsed.setter
    def IsIgnoreErrorsUsed(self, value: bool) -> None: ...
    @property
    def IsUseLightningTraceUsed(self) -> bool: ...
    @IsUseLightningTraceUsed.setter
    def IsUseLightningTraceUsed(self, value: bool) -> None: ...
    @property
    def IsCriterionSettingsUsed(self) -> bool: ...
    @IsCriterionSettingsUsed.setter
    def IsCriterionSettingsUsed(self, value: bool) -> None: ...
    @property
    def IsMinimumFluxUsed(self) -> bool: ...
    @IsMinimumFluxUsed.setter
    def IsMinimumFluxUsed(self, value: bool) -> None: ...
    @property
    def IsOverwriteUsed(self) -> bool: ...
    @IsOverwriteUsed.setter
    def IsOverwriteUsed(self, value: bool) -> None: ...
    @property
    def IsTargetSettingsUsed(self) -> bool: ...
    @IsTargetSettingsUsed.setter
    def IsTargetSettingsUsed(self, value: bool) -> None: ...
    @property
    def IsColorTargetsUsed(self) -> bool: ...
    @IsColorTargetsUsed.setter
    def IsColorTargetsUsed(self, value: bool) -> None: ...
    @property
    def IsResampleDetectorUsed(self) -> bool: ...
    @IsResampleDetectorUsed.setter
    def IsResampleDetectorUsed(self, value: bool) -> None: ...
    @property
    def ClearDetector(self) -> int: ...
    @ClearDetector.setter
    def ClearDetector(self, value: int) -> None: ...
    @property
    def Criterion(self) -> int: ...
    @Criterion.setter
    def Criterion(self, value: int) -> None: ...
    @property
    def UseSource(self) -> int: ...
    @UseSource.setter
    def UseSource(self, value: int) -> None: ...
    @property
    def UseDetector(self) -> int: ...
    @UseDetector.setter
    def UseDetector(self, value: int) -> None: ...
    @property
    def Boundary(self) -> int: ...
    @Boundary.setter
    def Boundary(self, value: int) -> None: ...
    @property
    def StartAt(self) -> int: ...
    @StartAt.setter
    def StartAt(self, value: int) -> None: ...
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @property
    def BitmapFile(self) -> int: ...
    @BitmapFile.setter
    def BitmapFile(self, value: int) -> None: ...
    @property
    def RaySampling(self) -> int: ...
    @RaySampling.setter
    def RaySampling(self, value: int) -> None: ...
    @property
    def EdgeSampling(self) -> int: ...
    @EdgeSampling.setter
    def EdgeSampling(self, value: int) -> None: ...
    @property
    def Target(self) -> float: ...
    @Target.setter
    def Target(self, value: float) -> None: ...
    @property
    def MinimumFlux(self) -> float: ...
    @MinimumFlux.setter
    def MinimumFlux(self, value: float) -> None: ...
    @property
    def OverallWeight(self) -> float: ...
    @OverallWeight.setter
    def OverallWeight(self, value: float) -> None: ...
    @property
    def TotalFlux(self) -> float: ...
    @TotalFlux.setter
    def TotalFlux(self, value: float) -> None: ...
    @property
    def NumberOfSources(self) -> int: ...
    @property
    def NumberOfDetectors(self) -> int: ...
    @property
    def NumberOfCriterion(self) -> int: ...
    @property
    def NumberOfConfigurations(self) -> int: ...
    @property
    def NumberOfRaySamplings(self) -> int: ...
    @property
    def NumberOfEdgeSamplings(self) -> int: ...
    @property
    def NumberOfBoundaries(self) -> int: ...
    def GetBoundaryAt(self, idx: int) -> str: ...
    def GetConfigurationAt(self, idx: int) -> str: ...
    def GetCriterionAt(self, idx: int) -> str: ...
    def GetDetectorAt(self, idx: int) -> str: ...
    def GetEdgeSamplingAt(self, idx: int) -> str: ...
    def GetRaySamplingAt(self, idx: int) -> str: ...
    def GetSourceAt(self, idx: int) -> str: ...

class ISEQOptimizationWizard(IWizard):
    @property
    def CommonSettings(self) -> IWizard: ...
    @property
    def IsSEQOptimizationWizard(self) -> bool: ...
    @property
    def IsDeleteVignetteUsed(self) -> bool: ...
    @IsDeleteVignetteUsed.setter
    def IsDeleteVignetteUsed(self, value: bool) -> None: ...
    @property
    def IsGlassUsed(self) -> bool: ...
    @IsGlassUsed.setter
    def IsGlassUsed(self, value: bool) -> None: ...
    @property
    def IsAirUsed(self) -> bool: ...
    @IsAirUsed.setter
    def IsAirUsed(self, value: bool) -> None: ...
    @property
    def IsAssumeAxialSymmetryUsed(self) -> bool: ...
    @IsAssumeAxialSymmetryUsed.setter
    def IsAssumeAxialSymmetryUsed(self, value: bool) -> None: ...
    @property
    def IsIgnoreLateralColorUsed(self) -> bool: ...
    @IsIgnoreLateralColorUsed.setter
    def IsIgnoreLateralColorUsed(self, value: bool) -> None: ...
    @property
    def IsAddFavoriteOperandsUsed(self) -> bool: ...
    @IsAddFavoriteOperandsUsed.setter
    def IsAddFavoriteOperandsUsed(self, value: bool) -> None: ...
    @property
    def IsRelativeXWeightUsed(self) -> bool: ...
    @IsRelativeXWeightUsed.setter
    def IsRelativeXWeightUsed(self, value: bool) -> None: ...
    @property
    def Type(self) -> int: ...
    @Type.setter
    def Type(self, value: int) -> None: ...
    @property
    def Data(self) -> int: ...
    @Data.setter
    def Data(self, value: int) -> None: ...
    @property
    def Reference(self) -> int: ...
    @Reference.setter
    def Reference(self, value: int) -> None: ...
    @property
    def PupilIntegrationMethod(self) -> int: ...
    @PupilIntegrationMethod.setter
    def PupilIntegrationMethod(self, value: int) -> None: ...
    @property
    def Ring(self) -> int: ...
    @Ring.setter
    def Ring(self, value: int) -> None: ...
    @property
    def Arm(self) -> int: ...
    @Arm.setter
    def Arm(self, value: int) -> None: ...
    @property
    def Grid(self) -> int: ...
    @Grid.setter
    def Grid(self, value: int) -> None: ...
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @property
    def StartAt(self) -> int: ...
    @StartAt.setter
    def StartAt(self, value: int) -> None: ...
    @property
    def Obscuration(self) -> float: ...
    @Obscuration.setter
    def Obscuration(self, value: float) -> None: ...
    @property
    def GlassMin(self) -> float: ...
    @GlassMin.setter
    def GlassMin(self, value: float) -> None: ...
    @property
    def GlassMax(self) -> float: ...
    @GlassMax.setter
    def GlassMax(self, value: float) -> None: ...
    @property
    def GlassEdge(self) -> float: ...
    @GlassEdge.setter
    def GlassEdge(self, value: float) -> None: ...
    @property
    def AirMin(self) -> float: ...
    @AirMin.setter
    def AirMin(self, value: float) -> None: ...
    @property
    def AirMax(self) -> float: ...
    @AirMax.setter
    def AirMax(self, value: float) -> None: ...
    @property
    def AirEdge(self) -> float: ...
    @AirEdge.setter
    def AirEdge(self, value: float) -> None: ...
    @property
    def RelativeXWeight(self) -> float: ...
    @RelativeXWeight.setter
    def RelativeXWeight(self, value: float) -> None: ...
    @property
    def OverallWeight(self) -> float: ...
    @OverallWeight.setter
    def OverallWeight(self, value: float) -> None: ...
    @property
    def NumberOfTypes(self) -> int: ...
    @property
    def NumberOfDataTypes(self) -> int: ...
    @property
    def NumberOfReferences(self) -> int: ...
    @property
    def NumberOfPupilIntegrationMethods(self) -> int: ...
    @property
    def NumberOfRings(self) -> int: ...
    @property
    def NumberOfArms(self) -> int: ...
    @property
    def NumberOfGrids(self) -> int: ...
    @property
    def NumberOfConfigurations(self) -> int: ...
    def GetArmAt(self, idx: int) -> str: ...
    def GetConfigurationAt(self, idx: int) -> str: ...
    def GetDataTypeAt(self, idx: int) -> str: ...
    def GetGridAt(self, idx: int) -> str: ...
    def GetPupilIntegrationMethodAt(self, idx: int) -> str: ...
    def GetReferenceAt(self, idx: int) -> str: ...
    def GetRingAt(self, idx: int) -> str: ...
    def GetTypeAt(self, idx: int) -> str: ...

class ISEQOptimizationWizard2(IWizard):
    @property
    def CommonSettings(self) -> IWizard: ...
    @property
    def Criterion(self) -> CriterionTypes: ...
    @Criterion.setter
    def Criterion(self, value: CriterionTypes) -> None: ...
    @property
    def SpatialFrequency(self) -> float: ...
    @SpatialFrequency.setter
    def SpatialFrequency(self, value: float) -> None: ...
    @property
    def XSWeight(self) -> float: ...
    @XSWeight.setter
    def XSWeight(self, value: float) -> None: ...
    @property
    def YTWeight(self) -> float: ...
    @YTWeight.setter
    def YTWeight(self, value: float) -> None: ...
    @property
    def Type(self) -> OptimizationTypes: ...
    @Type.setter
    def Type(self, value: OptimizationTypes) -> None: ...
    @property
    def Reference(self) -> ReferenceTypes: ...
    @Reference.setter
    def Reference(self, value: ReferenceTypes) -> None: ...
    @property
    def UseGaussianQuadrature(self) -> bool: ...
    @UseGaussianQuadrature.setter
    def UseGaussianQuadrature(self, value: bool) -> None: ...
    @property
    def UseRectangularArray(self) -> bool: ...
    @UseRectangularArray.setter
    def UseRectangularArray(self, value: bool) -> None: ...
    @property
    def Rings(self) -> int: ...
    @Rings.setter
    def Rings(self, value: int) -> None: ...
    @property
    def Arms(self) -> PupilArmsCount: ...
    @Arms.setter
    def Arms(self, value: PupilArmsCount) -> None: ...
    @property
    def Obscuration(self) -> float: ...
    @Obscuration.setter
    def Obscuration(self, value: float) -> None: ...
    @property
    def GridSizeNxN(self) -> int: ...
    @GridSizeNxN.setter
    def GridSizeNxN(self, value: int) -> None: ...
    @property
    def DeleteVignetted(self) -> bool: ...
    @DeleteVignetted.setter
    def DeleteVignetted(self, value: bool) -> None: ...
    @property
    def UseGlassBoundaryValues(self) -> bool: ...
    @UseGlassBoundaryValues.setter
    def UseGlassBoundaryValues(self, value: bool) -> None: ...
    @property
    def GlassMin(self) -> float: ...
    @GlassMin.setter
    def GlassMin(self, value: float) -> None: ...
    @property
    def GlassMax(self) -> float: ...
    @GlassMax.setter
    def GlassMax(self, value: float) -> None: ...
    @property
    def GlassEdgeThickness(self) -> float: ...
    @GlassEdgeThickness.setter
    def GlassEdgeThickness(self, value: float) -> None: ...
    @property
    def UseAirBoundaryValues(self) -> bool: ...
    @UseAirBoundaryValues.setter
    def UseAirBoundaryValues(self, value: bool) -> None: ...
    @property
    def AirMin(self) -> float: ...
    @AirMin.setter
    def AirMin(self, value: float) -> None: ...
    @property
    def AirMax(self) -> float: ...
    @AirMax.setter
    def AirMax(self, value: float) -> None: ...
    @property
    def AirEdgeThickness(self) -> float: ...
    @AirEdgeThickness.setter
    def AirEdgeThickness(self, value: float) -> None: ...
    @property
    def StartAt(self) -> int: ...
    @StartAt.setter
    def StartAt(self, value: int) -> None: ...
    @property
    def OverallWeight(self) -> float: ...
    @OverallWeight.setter
    def OverallWeight(self, value: float) -> None: ...
    @property
    def ConfigurationNumber(self) -> int: ...
    @ConfigurationNumber.setter
    def ConfigurationNumber(self, value: int) -> None: ...
    @property
    def UseAllConfigurations(self) -> bool: ...
    @UseAllConfigurations.setter
    def UseAllConfigurations(self, value: bool) -> None: ...
    @property
    def FieldNumber(self) -> int: ...
    @FieldNumber.setter
    def FieldNumber(self, value: int) -> None: ...
    @property
    def UseAllFields(self) -> bool: ...
    @UseAllFields.setter
    def UseAllFields(self, value: bool) -> None: ...
    @property
    def AssumeAxialSymmetry(self) -> bool: ...
    @AssumeAxialSymmetry.setter
    def AssumeAxialSymmetry(self, value: bool) -> None: ...
    @property
    def IgnoreLateralColor(self) -> bool: ...
    @IgnoreLateralColor.setter
    def IgnoreLateralColor(self, value: bool) -> None: ...
    @property
    def AddFavoriteOperands(self) -> bool: ...
    @AddFavoriteOperands.setter
    def AddFavoriteOperands(self, value: bool) -> None: ...
    @property
    def IsHighManufacturingYieldAvailable(self) -> bool: ...
    @property
    def OptimizeForBestNominalPerformance(self) -> bool: ...
    @OptimizeForBestNominalPerformance.setter
    def OptimizeForBestNominalPerformance(self, value: bool) -> None: ...
    @property
    def OptimizeForManufacturingYield(self) -> bool: ...
    @OptimizeForManufacturingYield.setter
    def OptimizeForManufacturingYield(self, value: bool) -> None: ...
    @property
    def ManufacturingYieldWeight(self) -> float: ...
    @ManufacturingYieldWeight.setter
    def ManufacturingYieldWeight(self, value: float) -> None: ...
    @property
    def UseMaximumDistortion(self) -> bool: ...
    @UseMaximumDistortion.setter
    def UseMaximumDistortion(self, value: bool) -> None: ...
    @property
    def MaxDistortionPct(self) -> float: ...
    @MaxDistortionPct.setter
    def MaxDistortionPct(self, value: float) -> None: ...

class ISEQToleranceWizard(IToleranceWizard, IWizard):
    @property
    def ToleranceSettings(self) -> IToleranceWizard: ...

class IToleranceWizard(IWizard):
    @property
    def CommonSettings(self) -> IWizard: ...
    @property
    def IsSEQToleranceWizard(self) -> bool: ...
    @property
    def IsSurfaceRadiusUsed(self) -> bool: ...
    @IsSurfaceRadiusUsed.setter
    def IsSurfaceRadiusUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceThicknessUsed(self) -> bool: ...
    @IsSurfaceThicknessUsed.setter
    def IsSurfaceThicknessUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceDecenterXUsed(self) -> bool: ...
    @IsSurfaceDecenterXUsed.setter
    def IsSurfaceDecenterXUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceDecenterYUsed(self) -> bool: ...
    @IsSurfaceDecenterYUsed.setter
    def IsSurfaceDecenterYUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceTiltXUsed(self) -> bool: ...
    @IsSurfaceTiltXUsed.setter
    def IsSurfaceTiltXUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceTiltYUsed(self) -> bool: ...
    @IsSurfaceTiltYUsed.setter
    def IsSurfaceTiltYUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceSandAIrregularityUsed(self) -> bool: ...
    @IsSurfaceSandAIrregularityUsed.setter
    def IsSurfaceSandAIrregularityUsed(self, value: bool) -> None: ...
    @property
    def IsSurfaceZernikeIrregularityUsed(self) -> bool: ...
    @IsSurfaceZernikeIrregularityUsed.setter
    def IsSurfaceZernikeIrregularityUsed(self, value: bool) -> None: ...
    @property
    def IsElementDecenterXUsed(self) -> bool: ...
    @IsElementDecenterXUsed.setter
    def IsElementDecenterXUsed(self, value: bool) -> None: ...
    @property
    def IsElementDecenterYUsed(self) -> bool: ...
    @IsElementDecenterYUsed.setter
    def IsElementDecenterYUsed(self, value: bool) -> None: ...
    @property
    def IsElementTiltXUsed(self) -> bool: ...
    @IsElementTiltXUsed.setter
    def IsElementTiltXUsed(self, value: bool) -> None: ...
    @property
    def IsElementTiltYUsed(self) -> bool: ...
    @IsElementTiltYUsed.setter
    def IsElementTiltYUsed(self, value: bool) -> None: ...
    @property
    def IsIndexUsed(self) -> bool: ...
    @IsIndexUsed.setter
    def IsIndexUsed(self, value: bool) -> None: ...
    @property
    def IsIndexAbbePercentageUsed(self) -> bool: ...
    @IsIndexAbbePercentageUsed.setter
    def IsIndexAbbePercentageUsed(self, value: bool) -> None: ...
    @property
    def IsFocusCompensationUsed(self) -> bool: ...
    @IsFocusCompensationUsed.setter
    def IsFocusCompensationUsed(self, value: bool) -> None: ...
    @property
    def SurfaceRadiusUnitType(self) -> DefaultAndFringes: ...
    @SurfaceRadiusUnitType.setter
    def SurfaceRadiusUnitType(self, value: DefaultAndFringes) -> None: ...
    @property
    def SurfaceTiltXUnitType(self) -> DefaultAndDegrees: ...
    @SurfaceTiltXUnitType.setter
    def SurfaceTiltXUnitType(self, value: DefaultAndDegrees) -> None: ...
    @property
    def SurfaceTiltYUnitType(self) -> DefaultAndDegrees: ...
    @SurfaceTiltYUnitType.setter
    def SurfaceTiltYUnitType(self, value: DefaultAndDegrees) -> None: ...
    @property
    def StartAt(self) -> int: ...
    @StartAt.setter
    def StartAt(self, value: int) -> None: ...
    @property
    def StartAtSurface(self) -> int: ...
    @StartAtSurface.setter
    def StartAtSurface(self, value: int) -> None: ...
    @property
    def StopAtSurface(self) -> int: ...
    @StopAtSurface.setter
    def StopAtSurface(self, value: int) -> None: ...
    @property
    def SurfaceRadius(self) -> float: ...
    @SurfaceRadius.setter
    def SurfaceRadius(self, value: float) -> None: ...
    @property
    def SurfaceRadiusFringes(self) -> float: ...
    @SurfaceRadiusFringes.setter
    def SurfaceRadiusFringes(self, value: float) -> None: ...
    @property
    def SurfaceThickness(self) -> float: ...
    @SurfaceThickness.setter
    def SurfaceThickness(self, value: float) -> None: ...
    @property
    def SurfaceDecenterX(self) -> float: ...
    @SurfaceDecenterX.setter
    def SurfaceDecenterX(self, value: float) -> None: ...
    @property
    def SurfaceDecenterY(self) -> float: ...
    @SurfaceDecenterY.setter
    def SurfaceDecenterY(self, value: float) -> None: ...
    @property
    def SurfaceTiltX(self) -> float: ...
    @SurfaceTiltX.setter
    def SurfaceTiltX(self, value: float) -> None: ...
    @property
    def SurfaceTiltXDegrees(self) -> float: ...
    @SurfaceTiltXDegrees.setter
    def SurfaceTiltXDegrees(self, value: float) -> None: ...
    @property
    def SurfaceTiltY(self) -> float: ...
    @SurfaceTiltY.setter
    def SurfaceTiltY(self, value: float) -> None: ...
    @property
    def SurfaceTiltYDegrees(self) -> float: ...
    @SurfaceTiltYDegrees.setter
    def SurfaceTiltYDegrees(self, value: float) -> None: ...
    @property
    def SurfaceSandAIrregularityFringes(self) -> float: ...
    @SurfaceSandAIrregularityFringes.setter
    def SurfaceSandAIrregularityFringes(self, value: float) -> None: ...
    @property
    def SurfaceZernikeIrregularityFringes(self) -> float: ...
    @SurfaceZernikeIrregularityFringes.setter
    def SurfaceZernikeIrregularityFringes(self, value: float) -> None: ...
    @property
    def ElementDecenterX(self) -> float: ...
    @ElementDecenterX.setter
    def ElementDecenterX(self, value: float) -> None: ...
    @property
    def ElementDecenterY(self) -> float: ...
    @ElementDecenterY.setter
    def ElementDecenterY(self, value: float) -> None: ...
    @property
    def ElementTiltXDegrees(self) -> float: ...
    @ElementTiltXDegrees.setter
    def ElementTiltXDegrees(self, value: float) -> None: ...
    @property
    def ElementTiltYDegrees(self) -> float: ...
    @ElementTiltYDegrees.setter
    def ElementTiltYDegrees(self, value: float) -> None: ...
    @property
    def Index(self) -> float: ...
    @Index.setter
    def Index(self, value: float) -> None: ...
    @property
    def IndexAbbePercentage(self) -> float: ...
    @IndexAbbePercentage.setter
    def IndexAbbePercentage(self, value: float) -> None: ...
    @property
    def TestWavelength(self) -> float: ...
    @TestWavelength.setter
    def TestWavelength(self, value: float) -> None: ...
    @property
    def SurfaceRadiusPercent(self) -> float: ...
    @SurfaceRadiusPercent.setter
    def SurfaceRadiusPercent(self, value: float) -> None: ...
    def SelectTolerancePreset(self, vendor: ToleranceVendor, grade: ToleranceGrade) -> None: ...

class IWizard:
    @property
    def Wizard(self) -> WizardType: ...
    def Apply(self) -> None: ...
    def Initialize(self) -> None: ...
    def LoadSettings(self) -> None: ...
    def OK(self) -> None: ...
    def ResetSettings(self) -> None: ...
    def SaveSettings(self) -> None: ...

class OptimizationTypes:
    RMS = 0
    PTV = 1

class PupilArmsCount:
    Arms_6 = 0
    Arms_8 = 1
    Arms_10 = 2
    Arms_12 = 3

class ReferenceTypes:
    Centroid = 0
    ChiefRay = 1
    Unreferenced = 2

class ToleranceGrade:
    Commercial = 0
    Precision = 1
    HighPrecision = 2
    CellPhoneLens = 3

class ToleranceVendor:
    Asphericon = 0
    EdmundOptics = 1
    Generic = 2
    LaCroix = 3
    Optimax = 4

class WizardType:
    NSCOptimization = 0
    NSCBitmap = 1
    NSCRoadwayLighting = 2
    SEQOptimization = 3
    NSCTolerance = 4
    SEQTolerance = 5



import hashlib
import json
from pathlib import Path

import pytest

from tests.config import CONFIG_DATA_FOLDER, REFERENCE_DATA_FOLDER, REFERENCE_VERSION
from zospy.analyses.base import AnalysisResult


def pytest_make_parametrize_id(config, val, argname):  # noqa: ARG001
    """Custom hook to control the name of dictionaries in the description."""
    if isinstance(val, dict):
        return hashlib.md5(json.dumps(val, sort_keys=True).encode("utf-8")).hexdigest()

    # Default ID for non-dictionary parameters
    return None


@pytest.fixture(scope="class")
def cfg_file(request) -> Path:
    # TODO check if we can programmatically write these files
    filename = f"{request.node.parent.name}-{request.node.name}"
    return request.config.rootpath / CONFIG_DATA_FOLDER / f"{filename}.CFG"


@pytest.fixture
def expected_data(request, optic_studio_version) -> AnalysisResult:
    filename = f"{optic_studio_version}-{request.fspath.basename}-{request.node.name}"

    data_file = request.config.rootpath / REFERENCE_DATA_FOLDER / f"{filename}.json"

    if not data_file.exists():
        pytest.skip(f"Data file {data_file} does not exist")

    return AnalysisResult.from_json(data_file.read_text(encoding="utf-8"))


@pytest.fixture
def reference_data(request) -> AnalysisResult:
    filename = f"{REFERENCE_VERSION}-{request.fspath.basename}-{request.node.name}"
    filename = filename.replace("matches_reference_data", "returns_correct_result")

    data_file = request.config.rootpath / REFERENCE_DATA_FOLDER / f"{filename}.json"

    if not data_file.exists():
        pytest.skip(f"Data file {data_file} does not exist")

    return AnalysisResult.from_json(data_file.read_text(encoding="utf-8"))



from __future__ import annotations

import hashlib
import json
import logging
import re
import sys
import traceback
from argparse import ArgumentParser
from operator import attrgetter
from pathlib import Path
from typing import Annotated, Any, Callable, Literal

import systems
import yaml
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError

import zospy as zp
from zospy.analyses.base import BaseAnalysisWrapper  # noqa: TCH001

logger = logging.getLogger("generate_test_reference_data")
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


TestParameters = dict[str, Any]


class VersionDependentParameters(BaseModel):
    opticstudio: str
    parameters: TestParameters = {}


class TestConfiguration(BaseModel):
    model: Callable[[zp.zpcore.OpticStudioSystem], zp.zpcore.OpticStudioSystem]
    analysis: type[BaseAnalysisWrapper]
    file: str
    test: str
    parameters: list[VersionDependentParameters | TestParameters] = [{}]
    parametrized: tuple[str, ...] | None = None

    @field_validator("analysis", mode="before")
    @classmethod
    def validate_analysis(cls, v: str):
        return attrgetter(v)(zp.analyses)

    @field_validator("model", mode="before")
    @classmethod
    def validate_model(cls, v: str):
        return getattr(systems, v)

    @field_validator("file")
    @classmethod
    def validate_filename(cls, v: str):
        assert v.startswith("test"), "file should start with 'test_'"
        assert v.endswith(".py"), "file should have extension '.py'"

        return v

    @field_validator("test")
    @classmethod
    def validate_testname(cls, v: str):
        assert v.startswith("test"), "test name should start with 'test_'"

        return v

    @model_validator(mode="after")
    @classmethod
    def validate_parametrized(cls, values: TestConfiguration):
        if len(parameters := values.parameters) > 1:
            parametrized = values.parametrized
            assert parametrized is not None, "Multiple sets of parameters are only allowed if parametrized is specified"

            for p in parameters:
                params = p.parameters if isinstance(p, VersionDependentParameters) else p

                if not all(p in params for p in parametrized):
                    raise PydanticCustomError(
                        error_type="invalid_parameters",
                        message_template=(
                            "All sets of parameters should contain the same keys as parametrized. Allowed keys: "
                            "{parametrized}, found keys: {params}"
                        ),
                        context={"parametrized": parametrized, "params": params.keys()},
                    )

        return values


class Configuration(BaseModel):
    redact_patterns: list[Annotated[re.Pattern, BeforeValidator(Configuration.validate_redact_patterns)]] = []
    tests: list[TestConfiguration]

    @staticmethod
    def validate_redact_patterns(v):
        regex = re.compile(v, re.MULTILINE | re.IGNORECASE)

        assert regex.groups == 0, (
            "Redaction patterns are not allowed to contain capturing groups. Use non-capturing "
            "groups (?:<characters>) instead."
        )

        return regex


AbsolutePath = Annotated[Path, AfterValidator(lambda p: p.resolve(strict=False))]


class CommandLineArguments(BaseModel):
    config: AbsolutePath
    connection_mode: Literal["standalone", "extension"]
    opticstudio_directory: AbsolutePath | None
    output_directory: Path


def _redact_json(json_string: str, patterns: list[re.Pattern]) -> tuple[str, list[str]]:
    matches = []

    for p in patterns:
        matches.extend(re.findall(p, json_string))
        json_string = re.sub(p, "REDACTED", json_string)

    return json_string, matches


def _map_parameter_names(parameters: TestParameters):
    """A function to create and apply a mapping for test parameters of the dict type to have consistent test naming.

    Mappings will be saved in 'dictionary_parameter_mapping.json' within the output directory.

    Parameters
    ----------
    parameters : dict
        A dictionary containing the test parameters in the form {param_name: param_value}.

    Returns
    -------
    dict
        A mapped parameters dictionary
    """
    if any(isinstance(value, dict) for value in parameters.values()):
        mapped_parameters = parameters.copy()

        for param_name, param in mapped_parameters.items():
            if isinstance(param, dict):
                mapped_name = hashlib.md5(json.dumps(param, sort_keys=True).encode("utf-8")).hexdigest()  # noqa: S324

                mapped_parameters[param_name] = mapped_name

        return mapped_parameters

    return parameters


def process_test(
    oss: zp.zpcore.OpticStudioSystem, test: TestConfiguration, args: CommandLineArguments, config: Configuration
):
    analysis = test.analysis
    test_file = test.file
    test_name = test.test

    test_parameters = test.parameters
    parametrized = test.parametrized

    optic_studio_version = str(oss.ZOS.version)

    for parameters in test_parameters:
        result_file_name = f"{optic_studio_version}-{test_file}-{test_name}"

        if isinstance(parameters, VersionDependentParameters):
            if not oss.ZOS.version.match(parameters.opticstudio):
                logger.info(f"Skipping {result_file_name} because it requires OpticStudio {parameters.opticstudio}")
                continue

            parameters = parameters.parameters  # noqa: PLW2901

        if parametrized:
            mapped_parameters = _map_parameter_names(parameters=parameters)

            result_file_name += f"[{'-'.join(str(mapped_parameters[p]) for p in parametrized)}]"

        output_json = args.output_directory / f"{result_file_name}.json"
        output_zos = args.output_directory / f"{result_file_name}.zmx"

        if output_json.exists():
            logger.info(f"Skipping {result_file_name} because it already exists")
            continue

        logger.info(f"Processing {result_file_name}")

        # Build model
        test.model(oss)

        try:
            # Run analysis
            result = analysis(**parameters).run(oss, oncomplete="Release")
        except Exception:
            logger.exception("Failed to run analysis %s. Traceback: %s", analysis.__name__, traceback.format_exc())
            continue

        try:
            result_json = result.to_json()
            result_json, redacted_strings = _redact_json(result_json, config.redact_patterns)

            if redacted_strings:
                logger.warning("\tRedacted strings: %s", "\n\t".join(redacted_strings))

            with open(output_json, "w", encoding="utf-8") as f:
                f.write(result_json)

            try:
                oss.save_as(str(output_zos.resolve()))
            except Exception:
                logger.exception("Failed to save Zemax file %s. Traceback: %s", output_zos, traceback.format_exc())
                output_zos.with_suffix(".txt").resolve().write_text("Failed to save OpticStudio model.")

                # Reconnect to OpticStudio to avoid potential issues
                zos = zp.ZOS(opticstudio_directory=args.opticstudio_directory)
                oss = zos.connect(mode=args.connection_mode)

        except TypeError:
            logger.exception("Failed to serialize %s. Traceback: %s", result_file_name, traceback.format_exc())


def main(args):
    args = CommandLineArguments.model_validate(args, from_attributes=True)

    if not args.config.exists():
        logger.fatal(
            f"Configuration file {args.config} does not exist. When running with the default arguments, "
            f"make sure you are running the script from the project root."
        )

        return 1

    if not args.output_directory.exists():
        logger.fatal(
            f"Output directory {args.output_directory} does not exist. When running with the default "
            f"arguments, make sure you are running the script from the project root."
        )

        return 1

    with open(args.config) as f:
        config = Configuration.model_validate(yaml.safe_load(f.read()))

    zos = zp.ZOS(opticstudio_directory=args.opticstudio_directory)
    oss = zos.connect(mode=args.connection_mode)

    if args.connection_mode == "extension":
        # Turn off UI updates for faster processing
        zos.Application.ShowChangesInUI = False

    logger.info(f"Generating test data files using OpticStudio {zos.version} and ZOSPy {zp.__version__}")

    for t in config.tests:
        process_test(oss, t, args, config)

    # Clean up Zemax workspace files
    logger.info("Removing Zemax workspace (.ZDA) files...")
    for f in args.output_directory.glob("*.ZDA"):
        f.unlink()

    return 0


argument_parser = ArgumentParser(description="Generate reference data files for ZOSPy unit tests.")
argument_parser.add_argument(
    "--config",
    type=Path,
    default=Path("scripts/generate_test_reference_data/tests.yaml"),
    help="Path to the configuration file.",
)
argument_parser.add_argument(
    "--connection-mode",
    choices=["standalone", "extension"],
    default="standalone",
    help="Connection mode to use with OpticStudio.",
)
argument_parser.add_argument(
    "--output-directory",
    type=Path,
    default=Path("tests/data/reference"),
    help="Path to the directory where the reference data files will be saved.",
)
argument_parser.add_argument(
    "--opticstudio-directory",
    type=Path,
    help="Path to the OpticStudio installation directory. This allows to choose a different version of OpticStudio.",
    required=False,
)

sys.exit(main(argument_parser.parse_args()))



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = "LMXF_Platforms"

class LMXF_Platforms(Enum):
    Unknown = 0
    SolidWorks = 1
    Creo = 2



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("ISAberrationTypes", "ISFlipTypes", "ISSamplings", "ISShowAsTypes")

class ISAberrationTypes:
    None_: ISAberrationTypes = None
    Geometric: ISAberrationTypes = None
    Diffraction: ISAberrationTypes = None

class ISFlipTypes:
    None_: ISFlipTypes = None
    TopBottom: ISFlipTypes = None
    LeftRight: ISFlipTypes = None
    TopBottomLeftRight: ISFlipTypes = None

class ISSamplings:
    None_: ISSamplings = None
    X2: ISSamplings = None
    X4: ISSamplings = None
    X8: ISSamplings = None
    X16: ISSamplings = None
    X32: ISSamplings = None
    X64: ISSamplings = None

class ISShowAsTypes:
    SimulatedImage: ISShowAsTypes = None
    SourceBitmap: ISShowAsTypes = None
    PSFGrid: ISShowAsTypes = None
